#! /usr/bin/node
/** @file       sa-worker          Command-line utility to join the Distributed Computer as a standalone worker
 *  @author     Matthew Palma, mpalma@sparc.network
 *              Ryan Rossiter, ryan@kingsds.network
 *  @date       March 2018, Sept, 2019
 *
 *  Exit codes:
 *  0  - Success
 *  1  - error, not otherwise specified
 *  12 - received dcpmsg.command.reload (by default - may be overridden in @dcpConfig)
 */

global.crypto = { getRandomValues: require('polyfill-crypto.getrandomvalues') }

require('dcp-rtlink').init();
require('dcp/config').load()
require('dcp/dcp-env').setPlatform('nodejs')

const process = require('process')
const fs = require('fs')
const os = require('os')
const path = require('path')
const wallet = require('dcp/wallet');
const debugging = require('dcp/debugging').scope('worker');

if (process.env.DCP_PROMISE_DEBUG) {
  require('promiseDebug').hook(undefined, true, (e) => {
    // Don't print errors when no work available
    return e.message === "No work available.";
  });
}

global.Worker = require('../lib/standaloneWorker').Worker
require('../lib/standaloneWorker').config.debug |= debugging()

let paymentAddr, idKeystore;

const options = {
  cores: require('os').cpus().length - 1,
  verbose: debugging()
}

function usage(errorlevel = 0) {
  let progName = path.basename(__filename);
  console.log(`
${progName} - Join the DCP Network as a Standalone Worker
Copyright (c) 2018-2019 Kings Distributed Systems Ltd., All Rights Reserved.

Usage:   ${progName} start [payment address/keystore] [options] |
         ${progName} benchmark

Global Options:
 --help          view this help
 --verbose       print extra debug output

Start Options:
 --idKeystorePath= specify the keystore to use when communicating
                 with the scheduler
 --scheduler=    specify the base URL for the scheduler
                 (default: ${dcpConfig.scheduler.location.href})
 --packages=     specify the base URL for the package server
 --cores=        specify the number of cores (sandboxes) to use for working (default: ${options.cores})
 --liveStatus    show the current status of the worker (default: false)
 --aggressive    the supervisor will check for slices at a faster interval,
                 this is used for testing where the slices should be completed ASAP

Note:            if no keystore is specified, we wil use the default account as
                 interpreted by wallet.get() (usually ~/.dcp/default.keystore)
Examples:
 Start Working using 3 cores and a newly created key:
   ${progName} start /home/me/dcp/default.keystore --cores=3
`)

  process.exit(errorlevel)
}

function fnExpand(a) {
  return a.replace(/^~\//, require("os").homedir + "/")
}

var dccEarned = 0.000
var completedSlices = 0;

function showPendingSlices (clearScreen = true, status = {}) {
  if (clearScreen) {
    process.stdout.write(String.fromCharCode(0x1b) + 'c')
  }
  
  console.log(
    "Pending Slices (" + completedSlices + " completed; "+ dccEarned.toFixed(8) +" earned)",
    (status.fetch  ? String.fromCharCode(0x2b8b) + ' ' : '') +
    (status.submit ? String.fromCharCode(0x2b89) + ' ' : '') +
    (status.error  ? String.fromCharCode(0x2bbf) : '')
  )
  console.log("-------------------------------------------------------------------------------")
  
  for (let w of compute.supervisor.workingWorkers) {
    const state = parseWorkingWorker(w, 'Sandbox ' + w.id.toString(16).padStart(4, '0').slice(-4));
    renderSandbox(state);
  }
  if (!compute.supervisor.workingWorkers.length) {
    console.log("No active workers")
  }
  console.log("-------------------------------------------------------------------------------")
  
  if (options.verbose) {
    const s = compute.supervisor
    console.log({
      readiedWorkers: s.readiedWorkers.length,
      initializedWorkers: s.initializedWorkers.length,
      workingWorkers: s.workingWorkers.length,
    })  

    console.log({
      queuedSlices: s.queuedSlices && s.queuedSlices.length || 0,
      workingSlices: s.workingSlices && s.workingSlices.length || 0,
    });

    console.log("-------------------------------------------------------------------------------")
  }

  console.log('Scheduler:        ' + dcpConfig.scheduler.location);
  console.log('Payment address:  0x' + paymentAddr.valueOf().replace(/^0x/, ''));
  console.log('Identity address: 0x' + protocol.keychain.currentAddress);

  console.log("-------------------------------------------------------------------------------")

  const now = new Date()
  const time = `${now.getHours().toString().padStart(2, '0')}:${now.getMinutes().toString().padStart(2, '0')}:${now.getSeconds().toString().padStart(2, '0')}`
  
  const s = os.uptime()
  const m = s / 60, h = m / 60, d = h / 24
  const days = Math.floor(d)
  const hhmm = `${Math.floor(h % 24).toString()}:${Math.floor(m % 60).toString().padStart(2, '0')}`
  const up = days ? `up ${days} days, ${hhmm}` : `up ${hhmm}`
  
  const loadavg = 'load average: ' + os.loadavg().map(e => e.toFixed(2)).join(', ')
  
  const free = (os.freemem() / (1024**3))
  const used = (os.totalmem() / (1024**3))
  // const mem = `${free}/${used} GB free`
  const mem = `${free >= 10 ? free.toFixed(0) : free.toFixed(1)} GB free`
  
  console.log(` ${time} ${up},  ${loadavg}, ${mem}`)
}

// Turn a record from supervisor.workingWorkers into a Sandbox State object
function parseWorkingWorker(worker, id) {
  return {
    id: 'Sandbox ' + worker.id.toString(16).padStart(4, '0').slice(-4),
    isStale: false, /// @todo: how to determine this?
    progress: worker.progress,
    isIndeterminate: worker.isIndeterminate,
    name: (worker.generator ? worker.generator.public.name : 'Anonymous job'),
    started: worker.started,
  }
}

// Turn a Sandbox State object into a row of text
function renderSandbox(sandbox) {
  // const scurries = ['\u257e\u2500', '\u257c\u2500', '\u2500\u257e', '\u2500\u257c']
  const scurries = ['\u2509\u2508\u2508','\u2508\u2509\u2508','\u2508\u2508\u2509']
  const scurry = scurries[Math.floor(Date.now() / 100) % scurries.length]
  
  const isStale = sandbox.isStale
  const staleFlag = isStale ? '\u274c' : ''
  const barLen = 28 - (staleFlag.length)
  const bar = ('\u2501'.repeat((sandbox.progress / 100) * barLen) + staleFlag)
        .padEnd(barLen, sandbox.isIndeterminate ? scurry : '\u2508')
    
  let pct = false;
  if (sandbox.isIndeterminate || !sandbox.progress || sandbox.progress === 100) {
    // show something suitable in the percent column
    pct = ' --- ';
    const t = Date.now() - sandbox.started;
    const mm = Math.floor(t / 60000);
    const ss = Math.floor(t / 1000) % 60;
    
    if (mm >= 100) {
      pct = mm.toString(10).padStart(4) + 'm';
    }
    else if (mm >= 1) {
      pct = mm.toString(10).padStart(2) + ':' + ss.toString(10).padStart(2, '0');
    }
    else {
      pct = ss.toString(10).padStart(4) + 's';
    }
  }
  else if (sandbox.progress === 100) {
    pct = '100 %';
  }
  
  if (!pct) {
    pct = `${(Number(sandbox.progress) || 0).toFixed(1).padStart(4)}%`;
  }
  
  console.log(
    col(sandbox.id, { width: 12 }),
    col(sandbox.name, { width: 31 }),
    bar,
    pct,
  )
}

function col(val, definition = { width: 8 }) {
  let str = String(val)
  const len = definition.width || 8
  
  if (typeof val === 'number') {
    str = val.toString().padStart(len)
  }
  
  str = str.padEnd(len).substr(0, len)
  
  return str
}

function handleNewDCPSandbox (sandbox) {
  const zeroes = '000';
  let key = `Sandbox 0x${(zeroes + sandbox.id.toString(16).toUpperCase()).slice(-zeroes.length)}`;
  let name = 'Ready to be initialized'
  
  console.log(key, '103: new DCP sandbox')
  
  sandbox.debug = debugging('sandbox');
  sandbox.debugLabel = 'sandbox-'+ key

  sandbox.addListener('sliceStart', function slice$Start (slice) {
    if (compute.workerDelayStartTime > 300) {
      compute.workerDelayStartTime -= 250
    }

    if (slice.job && slice.job.publicName) 
      name = slice.job.publicName;
    else if (sandbox.generator && sandbox.generator.public)
      name = sandbox.generator.public.name;

    if (options.verbose||!options.liveStatus) console.log(key, 'sliceStart - name:', name)
  })

  sandbox.addListener('sliceProgress', function slice$Progress (ev) {
    sandbox.isIndeterminate = ev.isIndeterminate;
  })

  sandbox.addListener('sliceFinish', function slice$Finish (ev) {
    if (typeof ev.result == 'undefined') return;
    completedSlices++;

    if (options.verbose||!options.liveStatus) console.log(key, 'sliceFinish - name:', name)
  })
  
  sandbox.addListener('workerStop', worker => {
    console.log(key, 'workerStop - generator address:',
      worker.generatorAddress ? worker.generatorAddress.substr(0, 10) : worker.generatorAddress);
  })
}

/** Start mining for DCC */
async function startMining(args, options) { 
  let sawSandbox

  sawSandbox = require('../lib/standaloneWorker')
  sawSandbox.config.debug = !!options.debug
  sawSandbox.config.debugLevel = 0

  if (options.debugLevel) {
    sawSandbox.config.debug = true
    sawSandbox.config.debugLevel = options.debugLevel
  }

  if (true || options.verbose) console.log(' * Starting Worker')

  // compute.addListener('sandboxStart', handleNewDCPSandbox)
  compute.addListener('sandboxStart', handleNewDCPSandbox)
  compute.addListener('dccCredit', ev => {
    try {
      console.log('dccCredit', ev.message.payment, dccEarned += parseFloat(ev.message.payment))
    }
    catch (error) {
      console.error('198:', error)
      console.log('     dccCredit', ev)
    }
  })
  
  // Status icons
  const status = {
    fetch: false,
    submit: false,
    error: 0,
    timeout: false,
  }
  
  compute.supervisor.addListener('submittingResult', ev => { status.submit = true; updateStatus() })
  compute.supervisor.addListener('submittedResult', ev => { status.submit = false; updateStatus() })
  compute.supervisor.addListener('submitResultFailed', ev => { status.submit = false; status.error++; updateStatus() })
  
  compute.supervisor.addListener('fetchingSlices', ev => { status.fetch = true; updateStatus() })
  compute.supervisor.addListener('fetchedSlices', ev => {
    status.fetch = false; //updateStatus();

    if (process.env.TEST_HARNESS) {
      process.send({
        request: 'fetchedSlices'
      })
    }
  })
  compute.supervisor.addListener('fetchSlicesFailed', ev => { status.fetch = false; status.error++; updateStatus() })

  /**
   * dcpmsg
   * 
   * Code for handling dcpmsg objects sent by the scheduler/protocol.
   * Used for sending broadcast messages to the client or executing a given command
   */
  this.dcpmsg = {lastMessage: ''}
  let that = this;
  protocol.on('dcpmsg', msg =>{
    console.log('246: dcpmsg', msg)
    switch (msg.message.type){
      //broadcast messages
      case 'broadcast':
        if (that.dcpmsg.lastMessage === msg.message.message){
          //That message has already been displayed, don't display again
        }
        else if (msg.message.message != null){
          //Display broadcast message
          that.dcpmsg.lastMessage = msg.message.message
          console.log( '\u{1f4ac} Broadcast message:', msg.message.message )
        }
        break

      //command messages
      case 'command':
        switch (msg.message.message.command) {
          case 'reload':        // Complete reload, with auto-login (in browser)
            console.log('* dcpmsg.command.reload; miner will reload')
            //In default config, this calls process.exit(0)
            eval(dcpConfig.standaloneWorker.reloadBehaviour)
            break
          
          case 'restart':       // Stop and restart workers
            console.log('* dcpmsg.command.restart; miner will restart')
            if (compute.mining){
              compute.stopMining()
              setImmediate(() => {
                compute.mine()
              });  
            }
            break
          
          case 'remove':        // Remove workers on a specific job ID
            console.log('* dcpmsg.command.remove; workers will be terminated')
            for (let worker of compute.supervisor.workingWorkers) {
              if (worker.generatorAddress === msg.message.message.genId) {
                compute.supervisor.returnWorker(worker, true)
              }
            }
           break
          
          default:
            console.warn(`WARN: Usupported dcpmsg command '${msg.message.message.command}'`)
        }
        break
      
      default:
        console.log(`WARN: Unsupported dcpmsg type '${msg.message.type}'`)
    }
  })
  
  function updateStatus() {
    if (status.error > 0) {
      setTimeout(() => {
        if (status.error > 0) {
          status.error--
        }
      }, 11)
    }
      
    if (options.liveStatus || !options.verbose || status.timeout)
      return;
    
    status.timeout = setTimeout(() => {
      if (status.fetch || status.submit || status.error)
        console.log(
          'Status:',
          status.fetch  ? String.fromCharCode(0x2b8b) + ' ' : '  ',
          status.submit ? String.fromCharCode(0x2b89) + ' ' : '  ',
          status.error  ? String.fromCharCode(0x2bbf) + ' ' : '  ',
          )
      status.timeout = false
    }, 10)
  }

  options.cores = parseInt(options.cores) || global.navigator.hardwareConcurrency || 1

  if (options.aggressive) {
    // check for slices every second
    compute.supervisor.setWatchdogInterval(1000);
  }

  console.log(" . Configured for scheduler " + dcpConfig.scheduler.location.href)
  console.log(` . Earned funds will be deposited in account ${paymentAddr} in bank ${dcpConfig.bank.location.href}`)
  console.log(` . Identity is ${idKeystore.address}`);
  
  /** XXX @todo remove keychain reference /wg,ke oct 2019 */
  protocol.keychain.addPrivateKey((await idKeystore.getPrivateKey()).valueOf(), true)
  compute.mine(options.cores, paymentAddr)

  // To let tests know we've actually started
  if (process.env.TEST_HARNESS) {
    process.send({
      request: 'Process Started'
    })
  }

  let liveInterval = false;
  if (options.liveStatus) {
    liveInterval = setInterval(() => showPendingSlices(true, status), 100);
  } else {
    // On a SIGUSR1, print the progress bars
    process.on('SIGUSR1', () => {
      showPendingSlices(false)
    })
  }

  return new Promise(function(resolve, reject) {
    process.on('SIGQUIT', () => {
      if (liveInterval)
        clearInterval(liveInterval);
      showPendingSlices(false);
      console.log('\n*** caught SIGQUIT; exiting...\n');
      compute.supervisor.stopWork(true);
      setTimeout(() => resolve(), 2000);
    })
  })
}

function runBenchmark (args, options) {
  let compute = new Compute()
  compute.addListener("benchmark-completed", function on$Benchmark$Complete (benchmarks) {
    console.log("Benchmarks Complete!")
    console.log("Results:", benchmarks)
  })

  return compute.benchmark()
}

/** Parse an argument vector, removing arguments which have been parsed, returning an options
 *  object. Arguments are key value parts. Arguments without a value are treated as true.
 *
 *  @param      arguments       The argument
 *  @param      array           An array of strings containing the list of arguments to parse
 *  @param      forceAll        If truey and not all arguments parsed, display an error and exit.
 *
 *  @returns    an options object
 */
function parseOptions (args, array, forceAll) {
  const options = {};

  for (let i = 0; i < args.length; i++) {
    if (args[i][0] !== '-' || args[i][1] !== '-') {
      continue
    }
    let equalPosition = args[i].indexOf('=')
    if(equalPosition === -1) {
      options[args[i].slice(2)] = true
      continue
    }
    let key = args[i].slice(2, equalPosition)
    let value = args[i].slice(equalPosition + 1)
    if (array.indexOf(key) === -1) {
      if (forceAll) {
        console.error("Unrecognized option:", key, value)
        usage(1)
      }

      continue
    }

    options[key] = value
  }

  // Do any options-related finessing...
  if (options.scheduler)
    dcpConfig.scheduler.location = new (require('dcp/dcp-url').URL)(options.scheduler)
  if (options.packages)
    dcpConfig.packageManager.location = new (require('dcp/dcp-url').URL)(options.packages)
                                       
  return options
}

function die(message) {
  console.log(message)
  process.exit()
}

/** Main program entry point */
async function main() {
  let argv = [].concat(process.argv).slice(1)
  let mode = argv[1]
  let privateKey

  if (argv.length < 2) {
    usage(0)
  }

  if (mode === 'start') {
    let paymentKeystore;
    let pksArg = argv[2];
    let tmpOptions = {};
    
    Object.assign(tmpOptions, parseOptions(argv, [ "key" ]));
    if (tmpOptions.hasOwnProperty('key')) {
      console.warn(' ! Warning: Using legacy --key option (preserved for CAC)');
      pksArg = tmpOptions.key;
    }

    if (typeof pksArg === 'string' && pksArg[0] != '-') {
      if (wallet.isAddress(pksArg)) {
        debugging('keystore') && console.log(`dcp-worker - Using payment address from command line`);
        paymentAddr = new wallet.Address(pksArg)
      }
      else if (wallet.isPrivateKey(pksArg)) {
        debugging('keystore') && console.log(`dcp-worker - Using private key from command line to generate payment address`);
        paymentKeystore = new wallet.Keystore(pksArg, false);
      }
      else {
        debugging('keystore') && console.log(`dcp-worker - Using payment address from file '${pksArg}'`)
        paymentKeystore = (await wallet.load(pksArg)).keystore;
      }
    } else {
      debugging('keystore') && console.log(`dcp-worker - Using payment address from wallet's default keystore`);
      paymentKeystore = (await wallet.get()).keystore;
    }
    
    if (paymentKeystore) {
      paymentAddr = paymentKeystore.address;
      wallet.add(paymentKeystore);
      console.log('*** using payment address as identity key', paymentKeystore.address);
      debugger
      wallet.addId(paymentKeystore); /** @todo XXX decouple id and bank keystores /wg dec 2019 */
    }
    else if (!paymentAddr) {
      die('Please specify address of the bank account in which to deposit funds earned')
    }

    debugging('keystore') && console.log(`dcp-worker - Payment address is ${paymentAddr}`);
  }
  debugging('platform') && console.log(`dcp-worker - platform is ${require('dcp/dcp-env').platform}`)
  argv = argv.slice(1)
  Object.assign(options, parseOptions(argv, [ "help", "scheduler", "packages", "idKeystorePath", "cores", "debug", "debugLevel", "liveStatus", "verbose", "force", "aggressive" ]));

  if (options.help)
    mode = "help"

  if (options.verbose) console.log(' * Loading Compute Bundle')
  require('dcp-client/dist/compute.min.js') // defer load as late as possible so config correct at load-time

  // Provide ability for parent process to spoof the
  // miner's capability object, used for testing
  if (process.env.TEST_HARNESS) {
    compute.supervisor.on('capabilitiesCalculated', (originalCaps) => {
      let reportedCaps = originalCaps;

      Object.defineProperty(compute.supervisor, 'capabilities', {
        get: () => reportedCaps
      });

      let setObjProps = (obj, source) => {
        for (let p in source) {
          if (typeof source[p] === 'object') setObjProps(obj[p], source[p]);
          else obj[p] = source[p];
        }
      }

      process.on('message', ({ request, data }) => {
        if (request === 'setCaps') {
          // Deep copy the original caps
          reportedCaps = JSON.parse(JSON.stringify(originalCaps));
          setObjProps(reportedCaps, data);
        }
        else if (request === 'resetCaps') reportedCaps = originalCaps;
        // else if (request === 'reportCaps') console.error(reportedCaps);
      });
    });
  }

  if (options.idKeystorePath) {
    idKeystore = (await wallet.load(options.idKeystorePath)).keystore
  } else {
    try {
      debugger
      idKeystore = await wallet.getId()
    } catch(e) {
      if (!e.code || e.code !== 'ENOENT')
        throw e;
      idKeystore = await new wallet.IdKeystore(null, '')
      console.log('\x07 ! *** Warning: generating dynamic identity: ', idKeystore.privateKey);
    }
  }
  
  switch(mode) {
    case "start":
      await startMining(argv.slice(1), options)
      break
    case "benchmark":
      await runBenchmark(argv.slice(1), options)
      break
    default:
      console.error("Invalid mode: " + mode)
    case "help":
      usage()
  }
}

main().finally(() => setImmediate(process.exit))
