/**
 * @file        config.js       <p>
 *                              Basic configuration module for DCP applications
 *                              which run on node.
 *                       </p><p>
 *                              Two files are loaded, which both alter global variable
 *                              'dcpConfig'. The first, dcp-default-config.js is the default
 *                              configuration, and is stored in the git repo. The
 *                              second, dcp-site-config.js is used to override any
 *                              settings in dcp-default-config.js for this particular
 *                              machine.  If the first file is missing, an exception
 *                              will be thrown, but a missing dcp-site-config is not an
 *                              error condition.
 *                       </p><p><blockquote>
 *                              *NOTE*: The dcp-config.js file used by the web applications is
 *                                      is not hand-written, but rather generated with the
 *                                      makeWebConfig.js program.  That program uses this library
 *                                      to find configuration values that are then published for
 *                                      web applications to use.
 *                       </blockquote></p><p>
 *                              We can also set override config variables with DCP software
 *                              running on node.  DCP_CONFIG_xyz=abc will override the
 *                              dcpConfig.xyz variable with the value abc.  If the value starts
 *                              with a double-quote, it will be passed through `JSON.parse()`.
 *                              An overlay algorithm which modifies only specified properties
 *                              will used for any JSON objects found in environment variables.
 *                       </p><p>
 *                              The dcp-default-config.js file is always loaded from ../etc as resolved 
 *                              by require from this module, unless overriden with environment
 *                              variable DCP_DEFAULT_CONFIG. This means that its location (../etc)
 *                              is mostly influenced by rtLink.js.
 *                       </p><p>
 *                              The dcp-site.config.js file may be loaded from several sources.
 *                              In order, the directories searched are:
 *                              <ul>
 *                               <li> {installLocation}/etc/dcp-site-config.js
 *                               <li> anything from `addConfigFile()`
 *                               <li> anything from the DCP_SITE_CONFIG_FILES environment variable,
 *                                 which is a colon-delimited list of files, which are resolved
 *                                 relative to the current directory
 *                               </ul>
 *                       </p><p>
 *                              Functions available inside config files:
 *                              <li> add(another site config filename)
 *                              <li> alias(path.to.config.value)
 *                       </p><p>
 *                              The dcp-default-config.js file is always formatted as arbitrary JS
 *                              code that is expected to define a global object named dcpConfig. 
 *                       </p><p>
 *                              The dcp-site-config.js file can be either an overlay object literal -
 *                              whose leaf properties override the same-named leaf properties in 
 *                              dcpConfig - or arbitrary JS code that is treated as a function body.
 *                              If the resultant function returns an object, it is treated as an
 *                              overlay object.
 *                       </p><p>
 *                              `alias()` declarations are processed after include files and environment
 *                              variables.  This means that aliases in the default config file will 
 *                              work as expected when the value that is aliased gets modified by a 
 *                              subsequent declartion, provided the override happens before `load()`
 *                              completes.
 *                       </p><p>
 *                              This library contains a 'configSections' constant. Its purpose is
 *                              to guarantee the existence of certain top-level objects so that we
 *                              don't have to test for them. It can be modified before the call to
 *                              `.load()` if it needs monkey-patching.
 * </p>
 *
 * @author      Wes Garland, wes@kingds.network
 * @date        July 2018
 * @see         ../etc/dcp-default-config.js
 * @module      config
 */

const rtLink = require('dcp-rtlink')
const path = require('path')
const fs = require('fs')
const vm = require('vm')
const configSections = [ 'board', 'modules', 'database', 'global', 'scheduler', 'needs' ]
const siteConfigFiles = [
  rtLink.installLocation + '/etc/local-config.json',
  rtLink.installLocation + '/etc/dcp-site-config.js',
] 
const loadedFiles = []
const configFuns = {
  require: require,
  path:    require('path'),
  fs:      require('fs'),
  process: require('process'),
  isWin:   process.platform === 'win32',
}  /**< Functions available from inside the configuration files */

/** Add an overlay object to the existing dcpConfig */
configFuns.add = function addOverlayObject (x) {
  return addConfig(global.dcpConfig, x)
}

/** Whatever the return value is assigned to becomes an alias for the target property.
 *
 *  @param     target    the.full.property.name, starting with an ownProperty of dcpConfig
 */
configFuns.alias = function alias (target) {
    return new Alias(target)
}

if (typeof global !== 'object') { global = GLOBAL /* old node */ } // eslint-disable-line

/** Constructor for alias properties, resolved with a mark-and-sweep type of algorithm
 *  that traverses the dcpConfig object graph, looking for instances of Alias.
 *
 *  @see patchupAlias
 */
function Alias(target) {
  this.target = target
  this.mark = 0
}

/** Configuration function to implement url('http://mysite.com:443') syntax, nominally
 *  for configuring IP addresses etc of system daemons.  Implemented with the dcp-url
 *  module.
 * 
 *  @param   urlString    
 *  @returns an instance of require('dcp-url').URL 
 */
configFuns.url = function url (urlString) {
  return new (require('dcp/dcp-url').URL)(urlString)
}

/** Traverse an object graph to find a node
 *  @param   obj          The object at the root of the graph
 *  @parma   configPath   A string identifying the path.to.the.node 
 *  @return  the node
 */
function resolveConfigValue(obj, configPath) {
  let entries = configPath.split('.')
  let entry
  for (entry = dcpConfig; entries.length; entry = entry[entries.shift()]) {
  }
  return entry
}

/* mark-and-sweep link resolver that handles links (aliases) including transitive links. 
 * Circulars links throw.
 *
 * @param       obj     {object}        Top of a directed (cyclic ok) graph which we search for Alias directives
 * @param       path    {string}        Label describing how we got 'here' from the top node of the top-most
 *                                      call on the recursion stack.
 *                                   
 */
function patchupAliases(obj, path) {
  if (process.env.DCP_DEBUG_CONFIG && obj === dcpConfig) { console.log(' * config patchup phase') }
  if (!patchupAliases.unique) {
    patchupAliases.unique = 1
  }
  if (!path) {
    path='dcpConfig'
  }
  for (prop in obj) {
    if (!obj.hasOwnProperty(prop)) {
      continue
    }
    patchupAliases.unique++
    while (typeof obj[prop] === "object" && obj[prop] instanceof Alias) {
      if (obj[prop].mark === patchupAliases.unique) {
        throw new Error(`Circular alias ${path}.${prop}`)
      } else {
        obj[prop].mark = patchupAliases.unique
      }
      if (String(process.env.DCP_DEBUG_CONFIG).match(/patchup/)) {
        console.log(' * config patchup ' + path + '.' + prop + ' <- ' + obj[prop].target + '\t('+ resolveConfigValue(dcpConfig, obj[prop].target) + ')')
      }
      obj[prop] = resolveConfigValue(dcpConfig, obj[prop].target)
    }
    if (typeof obj[prop] === "object") {
      patchupAliases(obj[prop], path + '.' + prop)
    }
  }
}
  
/** Load the DCP configuration file(s).  Should be invoked with no arguments,
 *  unless testing.
 *
 *  @param      filename        Alternate filename for ../etc/dcp-default-config.js
 *  @param      siteConfig      false to skip loading the site-specific files, string
 *                              to load a specific site-specific file, undefined
 *                              for normal behaviour (search for multiples)
 */
function load (filename, siteConfig) {
  var dcpConfig /* vars for use by eval code */
  var fn

  function indirectEval(code, filename, firstLineNumber) {
    return vm.runInThisContext(code, {filename: filename, lineOffset: firstLineNumber - 1})
  }

  function evalFile(filename) {
    return vm.runInThisContext('(' + fs.readFileSync(filename, 'utf-8') + ')', filename, 0) // eslint-disable-line
  }
  
  if (process.env.DCP_SITE_CONFIG_FILES) {
    var flist = process.env.DCP_SITE_CONFIG_FILES.split(':')
    for (let i = 0; i < flist.length; i++) {
      exports.addConfigFile(path.resolve(path.relative(process.cwd(), flist[i])))
    }
  }

  if (load.loaded) {
    throw new Error('DCP Configuration has already been loaded')
  } else {
    load.loaded = true
  }

  if (!load.filename) {
    load.filename = filename || process.env['DCP_DEFAULT_CONFIG'] || '../etc/dcp-default-config.js'
  }
  if (process.env.DCP_DEBUG_CONFIG) { console.log(' * loading', load.filename) }

  /* Load the default config */
  global.dcpConfig = {}
  try {
    let filename = require.resolve(load.filename)
    loadedFiles.push(filename) 
    fn = indirectEval(`
                      (function(${Object.keys(configFuns).join(',')}, isInstalled, installLocation, runningLocation) {
                        ${require('fs').readFileSync(filename, 'utf-8')}
                      })`, require.resolve(load.filename), 1)
    dcpConfig = fn.apply(null, Object.values(configFuns).concat(
      [path.resolve(rtLink.installLocation) === path.resolve(rtLink.runningLocation), rtLink.installLocation, rtLink.runningLocation]))
  } catch(e) {
    e.message += ' in ' + require.resolve(load.filename);
    throw e;
  }
  if (typeof dcpConfig === 'undefined') {
    throw new Error('Error evaluating ' + load.filename + ' (dcpConfig is undefined)')
  }
  for (let i = 0; i < configSections.length; i++) {
    if (!dcpConfig.hasOwnProperty(configSections[i])) {
      dcpConfig[configSections[i]] = {}
    }
  }

  if (!load.hasOwnProperty('siteConfigFiles')) {
    if (typeof siteConfig !== 'undefined') {
      load.siteConfigFiles = snapshot(siteConfigFiles || [])
    } else {
      load.siteConfigFiles = snapshot(siteConfigFiles)
    }
  }

  /* Load site config files in turn */
  load.siteConfigFiles.forEach(function loadSiteConfig (filename) {
    let o, fun
    
    try {
      filename = path.resolve(filename)
      if (process.env.DCP_DEBUG_CONFIG === 'verbose') {
        console.log(' * checking for site config at', filename)
      }
      let fileContents = fs.readFileSync(filename, 'utf-8');
      if (fileContents.replace(/(\/\*([\s\S]*?)\*\/)|(\/\/(.*)$)/gm /* comments */, '').match(/^\s*{/)) {
        /* object overlay syntax */
        fileContents = 'return (' + fileContents + ')'
      }

      /* Create an IIFE to inject symbols for use by config object */
      fun = indirectEval(`(function(dcpConfig,${Object.keys(configFuns).concat(Object.keys(dcpConfig)).join(',')}){${fileContents}})`, filename, 1)
      o = fun.apply(null, [dcpConfig].concat(Object.values(configFuns)).concat(Object.values(dcpConfig)))

      if (o)
        addConfig(dcpConfig, o)

      loadedFiles.push(filename)
      if (process.env.DCP_DEBUG_CONFIG) {
        console.log(' * loaded site config at', filename)
      }
    } catch (e) {
      if (process.env.DCP_DEBUG_CONFIG === 'verbose') {
        console.log(' * did not load site config from', filename, '-', e.message)
      }
      if (!e.message.match(/^ENOENT:/)) {
        loadedFiles.push(filename + ' [error]')
        e.message += ' in ' + filename
        throw e
      } else {
        loadedFiles.push(filename + ' [missing]')
      }
    }
  })

  for (let v in process.env) {
    if (!process.env.hasOwnProperty(v) || !v.match(/^DCP_CONFIG_/)) {
      continue
    }
    if (process.env[v][0] === '{') {
      let prop = fixCase(v.slice(11))
      if (typeof dcpConfig[prop] !== 'object') {
        dcpConfig[prop] = {}
      }
      addConfig(dcpConfig[prop], JSON.parse(process.env[v]))
    } else {
      if (typeof dcpConfig[prop] === "object") {
        throw new Error("Cannot override configuration property " + prop + " with a string (is an object)")
      }
      dcpConfig[prop] = process.env[v]
    }
  }

  global.dcpConfig = dcpConfig
  patchupAliases(dcpConfig)
  return dcpConfig
}

/** Merge a new configuration object on top of an existing one. The new object
 *  is overlaid on the existing object, so that properties specified in the 
 *  existing object graph overwrite, but unspecified edges are left alone.
 *
 * *note* -     We treat objects with constructors other than Function or Object
 *              as though they were primitive values, as they may contain internal 
 *              state that is not represented solely by their own properties
 * 
 * @param       {object}        existing        The top node of an object graph whose 
 *                                              edges may be replaced
 * @param       {object}        neo             The top node of an object graph whose
 *                                              edges describe the replacement
 */
function addConfig (existing, neo) {
  for (let prop in neo) {
    if (!neo.hasOwnProperty(prop)) { continue }
    if (typeof neo[prop] === 'object' && !Array.isArray(neo[prop]) && ['Function','Object'].includes(neo[prop].constructor.name)) {
      if (typeof existing[prop] === 'undefined') {
        existing[prop] = {}
      }
      addConfig(existing[prop], neo[prop])
    } else {
      existing[prop] = neo[prop]
    }
  }
}

/** Turn UGLY_STRING into uglyString */
function fixCase (ugly) {
  let camelCase = ugly.toLowerCase()
  let idx

  while ((idx = camelCase.indexOf('_')) !== -1) {
    camelCase = camelCase.slice(0, idx) + camelCase[idx + 1].toUpperCase() + camelCase.slice(idx + 2)
  }

  return camelCase
}

/** Returns a unique array made from a */
function unique (a) {
  return a.reduce(function (p, c) {
    if (p.indexOf(c) < 0) { p.push(c) }
    return p
  }, [])
};

/* Blank out an object graph without destroying the objects themselves, so that
 * we can repopulate re-specified primitive type values during a reload in a way
 * that will be recognized by outside code which is holding references to container
 * objects.
 *
 * @param       obj     {object}        Object to wipe
 *
 * *note* -     We do not consider objects with constructors other than Function or Object
 *              for this wipe, as they may contain internal state that is not represented 
 *              solely by their own properties
 */
function blank (obj) {
  for (let prop in obj) {
    if (!obj.hasOwnProperty(prop) || ['Function','Object'].includes(obj[prop].constructor.name))
      continue
    switch (typeof obj[prop]) {
      case 'object':
        if (obj[prop] === null) {
          delete obj[prop]
          break
        }
      case 'function': // eslint-disable-line
        blank(obj[prop])
        break
      default:
        delete obj[prop]
        break
    }
  }
}

/** Snapshot an object by serializing/deserializing (~deep clone)
 *  @param  obj  Object to snapshot
 *  @returns the new object 
 */
function snapshot (obj) {
  let v8 = require('v8')

  if (v8.serialize) { return v8.deserialize(v8.serialize(obj)) /* future node */ }

  return JSON.parse(JSON.stringify(obj))
}

/** *note* - experimental - intended for HUP handlers etc. Tricky part is foreign references.. */
function reload () {
  var dcpConfig
  let backup = global.dcpConfig
  delete load.loaded
  dcpConfig = load()
  blank(backup)
  addConfig(backup, dcpConfig)
  global.dcpConfig = backup
  patchupAliases(global.dcpConfig)
}

/** Traverse the dcpConfig object graph to find a node
 *  @param   {string}   path      A string identifying the path.to.the.node 
 *  @returns {object}   the node
 */
exports.getConfigValue = function config$$getConfigValue(path)
{
  return resolveConfigValue(dcpConfig, config)
}

exports.load = load
exports.reload = reload
exports.addConfig = addConfig
exports.addConfigFile = function (path) { siteConfigFiles.push(path) } /* experimental */
exports.configSections = configSections
exports.loadedFiles = loadedFiles
