/**
 * @file        env.js
 *              Module to return information about the DCP host environment
 *
 * @author      Wes Garland, wes@kingsds.network
 * @date        Oct 2019
 */
const { requireNative } = require('./webpack-native-bridge')
const debugging = require('./debugging').scope('dcp-env')
      
var platform = "unknown";
if (typeof __DCP_COMPUTE_MIN_JS__ === 'object')
  platform = requireNative ? 'nodejs' : 'vanilla-web'

exports.platformList = [
  "unknown",
  "nodejs",
  "bravojs",
  "cjs2-generic",
  "vanilla-web"
];

exports.browserPlatformList = [
  "bravojs",
  "vanilla-web"
]

Object.defineProperty(exports, 'platform', {
  get: function() {
    /** On initial load, the platform is 'unknown', unless you either
     *  1. have a working requireNative(), in which case we use requireNative('dcp/env.native').platform, or
     *  2. this script (webpack bundle) was loaded from a script tag whose dcp-env attribute was set
     */
    if (platform === 'unknown') {
      if (requireNative) {
        try {
          exports.setPlatform(requireNative('dcp/env-native').platform)
          debugging('set') && console.log(`dcp-env - platform '${exports.platform}' set via env-native`);
        } catch(e) {
          if (typeof document !== 'undefined' && typeof document.currentScript !== 'undefined') {
            exports.setPlatform((platform = document.currentScript.getAttribute('dcp-env') || platform))
          }
        }
      }
    }
    debugging('getter') && console.log(`dcp-env - platform is ${platform}`);
    return platform
  }
});

exports.setPlatform = function(newPlatform) {
  if (exports.platformList.indexOf(newPlatform) === -1) {
    throw new Error('Invalid platform: ' + newPlatform);
  }
  platform = newPlatform;
  exports.setPlatformDebugStack = new Error('\rPlatform set').stack  /* Breadcrumbs in case platform gets set wrong by somebody */
  exports.isBrowserPlatform = exports.browserPlatformList.indexOf(platform) !== -1
  
  return platform;
}

exports.setPlatformDebugStack = new Error('Never set')

/** Return a value based on the current platform
 *
 *  @param      cases           key-value-pair object where the key is the
 *                              name of the platform and the value is the value
 *                              to return.  An extra key, "default", is supported
 *                              in case the current platform does not match any
 *                              other key/
 *  @throws     {Error}         when the current platform is completely unhandled
 *                              (implies default key not provided)
 */
exports.switch = function env$switch(cases) {
  for (let prop in cases) {
    if (!cases.hasOwnProperty(prop))
      continue;

    if (prop === exports.platform) {
      return cases[prop]
    }
  }

  if (exports.platform === 'unknown') {
    console.warn('Warning: platform unknown in env module at', new Error().stack.split('\n')[2].replace(/^.*\(/,'').replace(/\) *$/,''))
  }
  
  if (cases.hasOwnProperty('default'))
    return cases.default

  throw new Error(`Platform ${exports.platform} not handled in cases object`)
}
