/**
 *  @file       dcp-url.js              Module for working with URLs; primarily intended for
 *                                      use in daemon configuration etc.  Works in Node 10 and
 *                                      current browsers. Probably works older Node, too.
 *
 *  @group      core-api
 *  @module     dcp-url
 *  @author     Wes Garland, wes@kingsds.network
 *  @date       May 2019
 */

/** Class for working with URLs; primarily intended for use in creating and communicating with
 *  DCP daemons; used by the configuration library to implement the url() function.
 *
 *  @constructor
 *  @param      url             A string describing a full URL, e.g. "http://localhost/".  Will be
 *                              parsed by Node's url.parse().
 *                              This object can be passed as an options argument to a NodeJS server
 *                              constructor, or used like 'window.location'.
 *
 *  @returns    {object}        An instance of exports.URL with at least the following properties:
 *                              <ul>
 *                              <li> hostname
 *                              <li> port
 *                              <li> protocol
 *                              <li> pathname
 *                              <li> href
 *                              </ul>
 *//** Class for working with URLs; primarily intended for use in creating and communicating with
 *  DCP daemons; used by the configuration library to implement the url() function.
 *
 *  @constructor
 *  @param      {object}        An instance of URL, or the return value of Node's url.parse
 *  @returns    {object}        An instance of exports.URL with at least the following properties:
 *                              <ul>
 *                              <li> hostname
 *                              <li> port
 *                              <li> protocol
 *                              <li> pathname
 *                              <li> href
 *                              </ul>
 */
exports.URL = function dcpUrl$$URL(url) {
  let parsedUrl
  
  if (typeof url === "object") {
    parsedUrl = url
  } else {
    parsedUrl = require('url').parse(url)
  }

  keys = [ 'hostname', 'port', 'protocol', 'pathname', 'href' ].concat(Object.keys(parsedUrl))
  keys = keys.filter(function unique(value, index, self) {
    return self.indexOf(value) === index
  })
  
  keys.forEach((key) => {
    if (parsedUrl[key] !== null)
      this[key] = parsedUrl[key]
  })

  if (!this.port && this.protocol.toLowerCase() === 'http:')
    this.port = 80;
  if (!this.port && this.protocol.toLowerCase() === 'https:')
    this.port = 443;

  switch (this.hostname.toLowerCase()) {
    case 'inaddr_any':
    case 'any/0':
    case 'any':
      this.hostname = 'inaddr_any'
      this.href = this.resolve('.')
    break
  }    
    
}

/** 
 *  @returns the full text version of URL 
 */
exports.URL.prototype.toString = function dcpURL$$URL$toString() {
  return this.href
}

/** 
 *  @returns the full text version of URL 
 */
exports.URL.prototype.valueOf = function dcpURL$$URL$valueOf() {
  return this.href
}

/** Form 1: Compare two URLs to see if they match origins per the Same-Origin Policy.
 *  @see https://developer.mozilla.org/en-US/docs/Web/Security/Same-origin_policy
 *
 *  @param      {string}        otherURL        The URL whose origin we are comparing to this one.
 *                                              It will be parsed using the same rules as strings
 *                                              passed into the exports.URL() constructor.
 *//** Form 2: Compare two URLs to see if they match origins per the Same-Origin Policy.
 *  @see https://developer.mozilla.org/en-US/docs/Web/Security/Same-origin_policy
 *
 *  @param      {object}        otherURL        The URL whose origin we are comparing to this one.
 *                                              Can be an instance of exports.URL() or a similar
 *                                              object, such as the result of Node url.parse(),
 *                                              browser's window.location, etc.
 */
exports.URL.prototype.sameOrigin = function dcpURL$$URL$sameOrigin(otherURL) {
  if (typeof otherURL === "string")
    otherURL = new (exports.URL)(otherURL)

  return this.protocol === otherURL.protocol && this.hostname === otherURL.hostname && this.port === otherURL.port
}
      
/** Resolves a target URL relative to the URL described by this instance in a manner similar to that
 *  of a Web browser resolving an anchor tag HREF.
 * 
 *  @param      {string}        path    The HREF URL being resolved
 *  @returns    {string}        The new URL string
 */
exports.URL.prototype.resolve = function dcpURL$$URL$resolve(path) {
  path = path.replace(/(^|\/)[.](\/[.])*(\/|$)/g, '/').replace(/[^/]+\/\.\.(\/|$)/g, '')  /* handle . and .. */
  return this.protocol + '//' + this.hostname + ':' + this.port + '/' + (this.pathname.replace(/\/[^/]*$/,'/') + path).replace(/\/\//g,'/').replace(/^\//,'')
}

/** Patch an object graph, transforming instances of URL into instances of exports.URL.
 *  Cyclic graphs are supported; depth limited to stack size. Edges are composed of
 *  own properties.
 * 
 *  @param      obj     {object}        The object graph to patch
 *  @param      seen    {Array}         An array of objects which should not be traversed 
 */
exports.patchup = function dcpURL$URL$patchObject(obj, seen) {
  if (!seen)
    seen = []
  seen.push(obj)

  Object.keys(obj).forEach((p) => {
    if ((typeof obj[p] !== 'object') || (obj[p] === null) || seen.indexOf(obj[p]) !== -1)
      return
    
    if (obj[p] instanceof URL)
      obj[p] = new (exports.URL)(obj[p])
    else
      exports.patchup(obj[p], seen)
  })
}


