/**
 * @file        dcp-xhr.js              XMLHttpRequest library for DCP, implemented via
 *                                      the xmlhttprequest-ssl NPM package, that transparently
 *                                      enables KeepAlive (for http and https) in a way that
 *                                      that is call-compatible with the browser code.
 *
 *                                      If this module is invoked in an environment where there
 *                                      is a self.XMLHttpRequest already defined (eg browser),
 *                                      it is assumed that that implementation behaves correctly
 *                                      and this module's export is replaced by that function.
 *
 * @module      dcp-xhr
 * @author      Wes Garland, wes@kingsds.network
 * @date        June 2019
 */

const { requireNative, process } = require('./webpack-native-bridge')
const debug = process.env.DCP_XHR_DEBUG

var vlog  
if (debug && debug.match(/\bverbose\b/)) {
  /** Verbose logging function: first argument should be an id number */
  vlog = function vlog() {
    var argv = Array.from(arguments)
    argv.unshift('DCP-XHR')
    if (typeof argv[1] === 'number')
      argv[1] += '>'
    else
      argv[0] += '>'
    console.log.apply(null, argv)
  }
}
else
  vlog = Function()

/** Return a node Agent (or workalike) object that matches the given URL.
 *  Agents are cached per-protocol so that xmlhttprequest-ssl can actually
 *  implement HTTP KeepAlive
 */
function getAgent(url) {
  let prot, agent
  let urlo = requireNative('url').parse(url)
  let pxHost = process.env.DCP_XHR_PROXY || (dcpConfig.network && dcpConfig.network.proxy)

  if (!getAgent.cache)
    getAgent.cache = {} /* one agent per protocol */

  prot = urlo.protocol.replace(/:/,'')
  if (prot != 'http' && prot != 'https')
    throw new Error('Protocol must be http: or https:!')

  if (getAgent.cache[prot])
    return getAgent.cache[prot]

  if (pxHost) {
    /* px support experimental! /wg jul 2019 */
    let pxAgent = requireNative(prot + '-proxy-agent')
    if (debug)
      console.log('* XHR using proxy', pxHost)
    agent = new pxAgent(pxHost)
  } else {
    agent = new (requireNative(prot).Agent)({
      keepAlive: true,
      keepAliveMsecs: (dcpConfig.network && dcpConfig.network.httpKeepAliveMsecs) || 180000
    })
  }

  getAgent.cache[prot] = agent
  return agent
}

/** 
 *  XMLHttpRequest constructor. Instanciates a Proxy which memoizes function calls and
 *  property writes before the open() method is invoked. Property accesses before open() 
 *  return values from a dummy instance.  
 *
 *  When open() is invoked, we examine the protocol in the URL and set up an instance of
 *  require('xmlhttprequest-ssl').XMLHttpRequest. We look at the pending call/writes memo,
 *  applying pending method calls and property writes to this object, and then we invoke
 *  its own open() method.
 *
 *  After open() is invoked, we proxy all property accesses to the 'real' xhr.
 *
 *  The reason for these gymnastics is a design error in the xmlhttprequest-ssl module; it
 *  requires knowledge of the Agent type, which depends on the protocol, in the constructor,
 *  which is contrary to the spec.; the protocol is only officially known once the URL is
 *  known.
 *
 *  @constructor
 *  @see xmlhttprequest-ssl
 *  @see https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest
 *  @returns   {object}     instance of Proxy
 */
exports.XMLHttpRequest = function dcpXhr$$XMLHttpRequest() {
  var thisId = exports.XMLHttpRequest.id = (exports.XMLHttpRequest.id || 0) + 1

  vlog(thisId, `Constructor`, new Error('VLOG XHR').stack.split('\n')[2])
  const XMLHttpRequest = requireNative('xmlhttprequest-ssl').XMLHttpRequest
  let dummy = new XMLHttpRequest() /* dummy request, used for type-testing */
  let pending = []                 /* list of pending function calls or property assignments, memoized before open() */
  let xhr                          /* actual XMLHttpRequest object; instanciated during open()  */

  function debugLabel(obj, prop) {
    if (typeof obj[prop] === 'function')
      return '[pxFunction' + (obj === xhr ? " " : " dummy ") + (obj[prop].name || prop) + ']'
    return obj[prop]
  }

  /* Apply the pending call/writes memo, then set up so that proxy will operate directly on xhr */
  let open = (function dcpXhr$$XMLHttpRequest$open(requestType, url) {
    vlog(thisId, `open`, arguments)
    xhr = new XMLHttpRequest({agent: getAgent(url)})
    pending.forEach((capture) => {
      if (capture.type === 'invoke') {
        xhr[capture.prop].apply(xhr, capture.arguments)
      } else {
        xhr[capture.prop] = capture.value
      }
    })

    dummy = pending = undefined
    xhr.open.apply(xhr, arguments)
    xhr.setDisableHeaderCheck(true)
    if (debug) console.log(' * XHR:', url)
    if (exports.referer)
      xhr.setRequestHeader('Referer', exports.referer)
  }).bind(this)

  return new Proxy({} /**< target */, {
    /* before .open() - return properties of dummy object
     * after  .open() - return properties of real object
     */
    get: function dcpXhr$$XMLHttpRequest$pxGetter(target, prop) {
      if (xhr) {
        vlog(thisId, `get ${prop}=${debugLabel(xhr, prop)}`)
        if (typeof xhr[prop] === 'function')
          return (function pxFunctionWrapper() {
            vlog(thisId, `invoke`, prop, arguments)
            return xhr[prop].apply(xhr, arguments)
          }).bind(xhr)
        return xhr[prop]
      }
      vlog(thisId, `get dummy ${prop}=${debugLabel(dummy, prop)}`)

      if (prop === 'open')
        return open
      if (typeof dummy[prop] === 'function') {
        return function dcpXhr$$XMLHttpRequest$captureFunctionCall() {
          if (xhr)
            xhr[prop]  /* don't memoize if .open() happened since property retrieval */
          else
            pending.push({type: 'invoke', prop: prop, arguments: Array.from(arguments)})
        }
      }
      else
        return dummy[prop]
    },

    /* memoize assignments or pass through to xhr if we're past open() */
    set: function dcpXhr$$XMLHttpRequest$pxSetter(target, prop, value) {
      if (xhr) {
        let old = debugLabel(xhr, prop)
        let neo
        xhr[prop] = value
        ne = debugLabel(xhr, prop)
        vlog(thisId, `set ${prop} ${old} -> ${neo}`)
      }
      else {
        pending.push({type: 'assignment', prop: prop, value: value})
        vlog(thisId, `memoize ${prop}=${value}`)
      }
      return true
    },

    has: function dcpXhr$$XMLHttpRequest$pxHas(target, key) {
      return key in (xhr || dummy)
    },

    ownKeys: function dcpXhr$$XMLHttpRequest$pxOwnKeys(target) {
      return Object.getOwnPropertyNames(xhr || dummy).concat(Object.getOwnPropertySymbols(xhr || dummy))
    },

    getOwnPropertyDescriptor: function dcpXhr$$XMLHttpRequest$pxGetOwnPropertyDescriptor(target) {
      return Object.getOwnPropertyDescriptor(xhr || dummy)
    },
  })
}

exports.referer = process.env.DCP_XHR_REFERER || (dcpConfig.network && dcpConfig.network['http-referer'])
if (!exports.referer) {
  if (module && module.parent && module.parent.filename) {
    exports.referer = requireNative('path').basename(module.parent.filename) + '/dcp-xhr'
  } else {
    /* webpack, probably */
    exports.referer = String(new Error().stack.split('\n')[1].match(/\(.*\)/)).replace(/^\(/,'').replace(/\)$/,'').replace(/\?.*/,'')
  }
}

switch(require('./dcp-env').platform) {
  case 'unknown':
    if (!window || !window.XMLHttpRequest)
      break
  case 'vanilla-web':  
  case 'bravojs':
    exports.XMLHttpRequest = window.XMLHttpRequest
}
