/**
 *  @file       debugging.js
 *              Utility functions for improving debug output, etc, in DCP JavaScript programs.
 *
 *  @author     Wes Garland, wes@kingsds.network
 *  @date       Nov 2019
 */

let selfDebugging = false;

/**
 * Declare a new debugging scope.  Each scope can watch a separate watchObject to determine
 * whether or not we are currently in debug mode.  Every scope watches the environment 
 * variable 'DCP_DEBUG'. 
 *
 * The debugging scope is queried by invoking the returned function, which has an optional
 * 'subsystem' argument.  When the query is made without the subsystem argument, the returned
 * function returns true if any subsystem in the scope is being debugged.  Otherwise, we return
 * true when the current debugging operations matches the debugging scope and specified subsystem.
 *
 * The watch object and the DCP_DEBUG environment variable are re-evaluated every time the 
 * returned function is called, however the results of matching for a given scope/subsystem are 
 * cached for speed.
 *
 * @param      {string}     scopeName      The name of the debugging scope (eg. 'compute')
 * @param      {object}     watchObject    An optional object whose `debug` property is watched
 *                                         to determine what debugging operation we are performing.
 * @returns    {function}   A function which returns boolean which is only true when the current
 *                          debugging operation describes both the current scope and subsystem.
 *
 * @example - logs when watchObject.debug or DCP_DEBUG='compute' or '*'
 * debugging = require('debugging').scope('compute');
 * debugging() && console.log('hello')
 *
 * @example - logs when watchObject.debug or DCP_DEBUG='compute::for' or '*'
 * debugging = require('debugging').scope('compute');
 * debugging('for') && console.log('hello')
 */
exports.scope = function debugging$$scope(scopeName, watchObject) {

  if (!scopeName)
    throw new Error('Must specify scopeName!');
  if (!watchObject)
    watchObject = {};

  selfDebugging && console.log(`debugging ${scopeName} - create scope`);

  function debugging(subsystem) {
    var re;
    let result = false;
    const env = (
      require('./dcp-env').isBrowserPlatform
        ? document && document.body && document.body.getAttribute('dcp-debug')
        : process.env.DCP_DEBUG
    );
    const debugSyms = (env || '') + (env && watchObject.debug ? ',' : '') + (watchObject.debug || '');

    if (debugging.cacheValidity !== debugSyms) {
      debugging.cache = {};
      debugging.cacheValidity = debugSyms;
      selfDebugging && console.log(`debugging ${scopeName} - initialized lookup cache`);
    }
    if (debugging.cache[subsystem])
      return debugging.cache[subsystem];

    /* No self-debugging code above to keep cache-hit and watch-object paths fast */
    
    loop: for (let debugSym of debugSyms.split(',')) {
      selfDebugging && console.log(`debugging ${scopeName} - examining debugSym ${debugSym}`)
      
      if (typeof subsystem === 'undefined') {
        if (typeof debugSym === 'boolean' && debugSym) {
          result = debugSym;
          selfDebugging && console.log(`debugging ${scopeName} - matched on boolean debugSym`);
          break;
        }
        if (debugSym === scopeName) {
          selfDebugging && console.log(`debugging ${scopeName} - matched on undecorated scope name`);
          result = true;
          break;
        }
      }

      switch(debugSym)
      {
        case '*':                               /* everything */
        case 'all':
        case '*:*':
        case 'all:*':
        case '*:all':
        case 'all:all':
          if (scopeName !== 'debugging') {
            selfDebugging && console.log(`debugging ${scopeName} - ${debugSym} matched everything`);
            result = true;
          }
          break loop;
        case scopeName + '::all' :              /* everything for this scope */
          selfDebugging && console.log(`debugging ${scopeName} - ${debugSym} matched everything for this scope`);
          result = true;
          break loop;
        case scopeName + '::*' :                
        case '*::' + subsystem:                 /* everything for this subsystem */
          selfDebugging && console.log(`debugging - ${debugSym} matched everything for this subsystem`);
          result = true;
          break loop;
        case 'all::' + subsystem:
        case scopeName + '::' + subsystem:      /* things for this scope and subsystem */
          selfDebugging && console.log(`debugging - ${debugSym} matched scope and subsystem`);
          result = true;
          break loop;
      }

      re = new RegExp('^(' + debugSym + ')$');
      selfDebugging && console.log(`debugging ${scopeName} - checking regular expression ${re}`);
      if (re.test(scopeName)) {
        selfDebugging && console.log(`debugging ${scopeName} - ${debugSym} matched regular expression for scope`);
        result = true;
        break loop;
      }
      else if (re.test(subsystem)) {
        selfDebugging && console.log(`debugging ${scopeName} - ${debugSym} matched regular expression for subsystem`);
        result = true;
        break loop;
      }
      else if (re.test(scopeName + '::' + subsystem)) {
        selfDebugging && console.log(`debugging ${scopeName} - ${debugSym} matched regular expression for subsystem`);
        result = true;
        break loop;
      }
      selfDebugging && console.log(`debugging ${scopeName} - ${debugSym} => no match in scope ${scopeName}`);
    }
    
    debugging.cache[subsystem] = result;
    return result;
  }

  return debugging;
}

selfDebugging = exports.scope('debugging')();
