"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = exports.UncaughtExceptionError = exports.SliceTooSlowError = exports.NoProgressError = void 0;

var _protocol = _interopRequireDefault(require("./protocol.js"));

var _eventEmitter = _interopRequireDefault(require("./eventEmitter.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

const protocol = typeof window !== 'undefined' && window.protocol ? window.protocol : new _protocol.default();
protocol.debugInfo = {
  version: '__GIT_VERSION__',
  branch: '__GIT_BRANCH__',
  timestamp: '__TIMESTAMP__'
};
const workerDependencies = ['/bravojs/bravo.js', '/worker.js']; // Worker states

const UNREADY = 'UNREADY'; // No Worker (web worker, saworker, etc) has been constructed yet

const READYING = 'READYING'; // Worker is being constructed and worker environment (bravojs, worker) is being set up

const READY = 'READY'; // Worker is ready and may or may not be initialized

const WORKING = 'WORKING'; // Worker is working

const TERMINATED = 'TERMINATED';

class NoProgressError extends Error {
  constructor(...args) {
    super(...args);

    _defineProperty(this, "errorCode", 'ENOPROGRESS');
  }

}

exports.NoProgressError = NoProgressError;

class SliceTooSlowError extends Error {
  constructor(...args) {
    super(...args);

    _defineProperty(this, "errorCode", 'ESLICETOOSLOW');
  }

}

exports.SliceTooSlowError = SliceTooSlowError;

class UncaughtExceptionError extends Error {
  constructor(...args) {
    super(...args);

    _defineProperty(this, "errorCode", 'ETOOMANYERRORS');
  }

}

exports.UncaughtExceptionError = UncaughtExceptionError;

class DistributedWorker extends _eventEmitter.default {
  /**
   * A DistributedWorker (ie. a worker sandbox) which executes distributed slices
   *
   * @constructor
   * @param {Worker} [WorkerConstructor = Worker] - Constructor for the "web" Worker
   * @param {object} [capabilities = {}] - the capabilites of this worker sandbox
   */
  constructor(WorkerConstructor = Worker, capabilities = {}) {
    super('DistributedWorker');
    this.WorkerConstructor = WorkerConstructor;
    this.capabilities = capabilities;
    this.generatorAddress = null;
    this.worker = null;
    this.ee = new _eventEmitter.default('DistributedWorkerInternal');
    this._state = UNREADY;
    this.progress = 100;
    this.progressTimeout = null;
    this.sliceTimeout = null;
    this.started = null;
    this.lastRun = null;
    this.requiresGPU = false;
    this.supervisor = typeof compute !== 'undefined' && compute.supervisor || null;
    this.packageURL = dcpConfig.packageManager.location;
    this.id = DistributedWorker.getNewId();
  }

  static getNewId() {
    return DistributedWorker.idCounter++;
  }

  get state() {
    return this._state;
  }

  set state(value) {
    if (DistributedWorker.debugState) {
      console.log(`XXX changing state of ${this.id}... ${this._state} -> ${value}`);
    }

    if (this.state === TERMINATED && value !== TERMINATED) {
      // For safety!
      throw new Error(`DistributedWorker set state violation, attepted to change state from ${this.state} to ${value}`);
    }

    return this._state = value;
  }

  get isReady() {
    return this.state === READY;
  }

  get isInitialized() {
    return !!this.generatorAddress;
  }

  get isWorking() {
    return this.state === WORKING;
  }

  get isTerminated() {
    return this.state === TERMINATED;
  }
  /**
   * Fetch the workerBootstrap file that will be used to initialize the worker.
   * This function by default makes a request to fetch the file. Override or
   * or monkey patch if you want to fetch it through other means.
   *
   * returns {Promise} - resolves to the workerBootstrap file string on success, rejects otherwise
   */


  getBootstrap() {
    let code;
    let url = this.packageURL.resolve('workerBootstrap.js');
    if (!this._getBootstrapCache) this._getBootstrapCache = {
      last: 0
    };
    if (code && Date().now - this._getBootstrapCache.last < dcpConfig.worker.bootstrapCacheTTL) return this._getBootstrapCache.code;
    code = protocol.justFetch(url, 'string', 'GET', true);
    this._getBootstrapCache.code = code;
    return code;
  }
  /**
   * What to do with the worker bootstrap string before passing it into the worker constructor.
   * You can override or monkey patch this method if not passing into a WebWorker.
   * By default it blobifies the bootstrap string and creates a URL Object with it.
   *
   * @param {string} bootstrapString - the string to be transformed
   * @returns Object with the following properties:<ul>
   * <li>           marshalledBootstrap - newly created url from URL.createObjectURL
   * <li>           cleanupFunction - optional function if marshalling bootstrap requires cleanup after worker starts
   * </ul>
   */


  marshallBootstrap(bootstrapString) {
    if (require('./dcp-env').platform === 'nodejs') {
      return [bootstrapString];
    } else if (navigator && navigator.userAgent && navigator.userAgent.match(/Version\//) && navigator.userAgent.match(/Safari\//)) {
      // trap Safari and try to blob-wrap the worker
      const blob = window.URL.createObjectURL(new Blob([bootstrapString], {
        type: 'application/javascript'
      }));

      const cleanup = () => {
        window.URL.revokeObjectURL(blob);
      };

      return [blob, cleanup];
    } else {
      return ['data:application/javascript;base64,' + btoa(bootstrapString)];
    }
  }
  /**
   * Fetches all the required dependencies the worker needs before it can
   * start executing slices. See @workerDependencies for list of url's fetched.
   *
   * @returns {string} - one string of all the dependencies concatenated in order
   */


  async getWorkerDependencies() {
    let promises, dependencies, code, url;
    promises = [];

    for (let dependency of workerDependencies) {
      url = this.packageURL.resolve(dependency);
      promises.push(protocol.justFetch(url, 'string', 'GET', true));
    }

    dependencies = await Promise.all(promises);
    code = '';

    for (let dependency of dependencies) {
      code += dependency;
    }

    return code;
  }
  /**
   * Readies the worker. This will result in the worker being ready and not initialized,
   * it will need to be initialized with a generator before it is able to do work.
   * 
   * @todo maybe preload specific modules or let the supervisor pass in what modules to load?
   * @throws on failure to ready
   */


  async start(delay = 0) {
    let bootstrap, bootstrapCleanup, dependencies;
    this.started = Date.now();
    this.state = READYING;
    if (delay > 0) await new Promise(resolve => setTimeout(resolve, delay));

    try {
      bootstrap = await this.getBootstrap();
      [bootstrap, bootstrapCleanup] = await this.marshallBootstrap(bootstrap);
      this.worker = new this.WorkerConstructor(bootstrap, {
        name: `DCP Sandbox #${this.id}`
      });
      this.worker.onmessage = this.onmessage.bind(this);
      this.worker.onerror = this.onerror.bind(this);
      if (typeof bootstrapCleanup === 'function') bootstrapCleanup();
      dependencies = await this.getWorkerDependencies();
      await this.eval(dependencies, 'workerDependencies.dcp');
      this.state = READY;
      this.emit('workerReadied', this);
      this.emit('sandboxStart', this);
    } catch (error) {
      console.error('Failed to start the worker', error);
      this.terminate(false);
      throw error;
    }
  }
  /**
   * This will initialize the worker with a generator, loading its worker code
   * into the sandbox.
   * 
   * @param {string} generatorAddress The address of the generator to initialize with
   * @throws on initialization failure
   */


  async initialize(generatorAddress) {
    try {
      // Setting the generatorAddress is initialization
      this.generatorAddress = generatorAddress;
      this.generator = this.job = await this.supervisor.fetchGenerator(this.generatorAddress); // Extract public data from generator, with defaults

      const {
        name: publicName = `Anonymous Job ${this.generator.address.slice(-6)}`,
        description: publicDescription = 'Discreetly helping make the world smarter.',
        link: publicLink = ''
      } = this.generator.public || {};
      this.public = {
        publicName,
        publicDescription,
        publicLink
      }; // Future: We may want other filename tags for appliances // RR Nov 2019

      await this.eval(this.generator.worker, `generator::${publicName}`);
      await this.applyWorkerRequirements(this.generator.requirements);
    } catch (e) {
      // Call terminate to clean up worker and set state to CRASHED
      this.terminate(false);
      throw e;
    }
  }
  /**
   * Evaluates a string inside the worker.
   *
   * @param {string} code - the code to evaluate in the worker
   * @param {string} filename - the name of the 'file' to help with debugging, no longer working though?
   * @returns {Promise} - resolves with eval result on success, rejects otherwise
   */


  eval(code, filename) {
    return new Promise((resolve, reject) => {
      let resultEvent, onSuccess, onFail;
      let msg = {
        request: 'eval',
        data: code,
        filename
      }; // To help denote wtf is going on

      switch (this.state) {
        case READYING:
          resultEvent = 'evalResult';
          break;

        case READY:
          resultEvent = 'mainModuleEvaluated';
          break;

        default:
          resultEvent = 'evalResult';
          break;
      } // I feel like this is easier to understand compared to just reassigning this.resolve and this.reject


      onSuccess = this.ee.once(resultEvent, function dw$$eval$successListener(event) {
        this.ee.removeListener('reject', onFail);
        resolve(event.data.data);
      }.bind(this));
      onFail = this.ee.once('reject', function dw$eval$failListener(event) {
        this.ee.removeListener(resultEvent, onSuccess);
        reject(event);
      }.bind(this));
      this.worker.postMessage(msg);
    });
  }
  /**
   * Resets the state of the worker bootstrap, without resetting the worker function if initialized.
   * Mostly used to reset the progress status before reusing a worker on another slice.
   * Must be called after @start.
   *
   * @returns {Promise} - resolves with result on success, rejects otherwise
   */


  resetWorkerState() {
    return new Promise((resolve, reject) => {
      let successCb, failTimeout;
      let msg = {
        request: 'resetState'
      };
      successCb = this.ee.once('resetStateDone', function dw$resetState$cb() {
        clearTimeout(failTimeout);
        resolve();
      });
      failTimeout = setTimeout(() => {
        this.ee.removeListener('resetStateDone', successCb);
        this.terminate(false);
        reject(new Error('resetState never received resetStateDone event from worker'));
      }, 3000);
      this.worker.postMessage(msg);
    });
  }

  checkWorkerCapabilities() {
    return new Promise((resolve, reject) => {
      let successCb, failTimeout;
      let msg = {
        request: 'checkCapabilities'
      };
      successCb = this.ee.once('checkCapabilitiesDone', function dw$workerCapabilities$cb(event) {
        clearTimeout(failTimeout);
        resolve(event.data.data);
      });
      failTimeout = setTimeout(() => {
        this.ee.removeListener('checkCapabilitiesDone', successCb);
        this.terminate(false);
        reject(new Error('checkCapabilities never received checkCapabilitiesDone event from worker'));
      }, 3000);
      this.worker.postMessage(msg);
    });
  }
  /**
   * Passes the generator's requirements object into the sandbox so that the
   * global access lists can be updated accordingly.
   *  e.g. disallow access to OffscreenCanvas without environment.offscreenCanvas=true present
   * Must be called after @start.
   *
   * @returns {Promise} - resolves with result on success, rejects otherwise
   */


  applyWorkerRequirements(requirements) {
    return new Promise((resolve, reject) => {
      let successCb, failTimeout;
      let msg = {
        request: 'applyRequirements',
        requirements
      };
      successCb = this.ee.once('applyRequirementsDone', function dw$applyRequirements$cb() {
        clearTimeout(failTimeout);
        resolve();
      });
      failTimeout = setTimeout(() => {
        this.ee.removeListener('applyRequirementsDone', successCb);
        this.terminate(false);
        reject(new Error('applyRequirements never received applyRequirementsDone event from worker'));
      }, 3000);
      this.worker.postMessage(msg);
    });
  }
  /**
   * Executes a slice received from the supervisor.
   * Must be called after @start.
   *
   * @param {object} slice - bare minimum data required for the generator/generator code to be executed on
   * @param {number} [delay = 0] the delay that this method should wait before beginning work, used to avoid starting all workers at once
   *
   * @returns {Promise} - resolves with result on success, rejects otherwise
   */


  async run(slice, delay = 0) {
    if (!this.isReady) {
      throw new Error("DistributedWorker.run: Worker is not ready to work, state=" + this.state);
    }

    this.state = WORKING;
    this.slice = slice; // Now wait for the delay if provided, prevents many workers starting at once from crashing the supervisor

    if (delay > 0) await new Promise(resolve => setTimeout(resolve, delay));
    if (!this.isWorking) return; // worker.terminate could have been called during the delay timeout
    // Prepare the worker to begin work

    if (!this.isInitialized) {
      await this.initialize(slice.generatorAddress);
    } else {
      if (this.generatorAddress !== slice.generatorAddress) {
        throw new Error(`DistributedWorker.run: Worker is already initialized and generator address doesn't match previous (${this.generatorAddress} !== ${generator.address})`);
      }

      await this.resetWorkerState();
    }

    let sliceHnd = {
      job: this.public,
      worker: this
    };
    this.resetProgressTimeout();
    this.resetSliceTimeout();
    return new Promise((resolve, reject) => {
      let onSuccess, onFail;
      onSuccess = this.ee.once('resolve', function dw$$eval$successListener(event) {
        this.ee.removeListener('reject', onFail);
        resolve(event);
      }.bind(this));
      onFail = this.ee.once('reject', function dw$eval$failListener(event) {
        this.ee.removeListener('complete', onSuccess);
        reject(event);
      }.bind(this));
      this.lastRun = Date.now();
      this.progress = null;
      this.time = [[0, performance.now()]];
      this.resetProgressTimeout();
      this.resetSliceTimeout();
      this.emit('sliceStart', sliceHnd); // Try to only pass the data required - keeps the message size low
      // and prevents worker code from snooping on internal data

      this.worker.postMessage({
        request: this.slice.request,
        data: this.slice.data
      });
    }).then(event => {
      // This is the complete callback when the slice completes
      this.state = READY; // TODO: Should sliceHnd just be replaced with this.public?

      this.emit('slice', sliceHnd);
      this.emit('sliceFinish', event.data);
      delete this.slice;
      return event.data;
    }).catch(event => {
      // This is the reject callback for when the slice throws an error
      this.terminate(false);
      this.emit('error', event, 'slice');
      this.emit('sliceError', sliceHnd, event);
      throw event;
    }).finally(() => {
      this.emit('sliceEnd', sliceHnd);
    });
  }

  resetProgressTimeout() {
    if (this.progressTimeout) clearTimeout(this.progressTimeout);
    this.progressTimeout = setTimeout(() => {
      if (DistributedWorker.debugWork) return console.warn("DistributedWorker.debugWork: Ignoring progress timeout");
      this.ee.emit('reject', new NoProgressError(`No progress event was received in the last ${dcpConfig.worker.progressTimeout / 1000} seconds.`));
    }, dcpConfig.worker.progressTimeout);
  }

  resetSliceTimeout() {
    if (this.sliceTimeout) clearTimeout(this.sliceTimeout);
    this.sliceTimeout = setTimeout(() => {
      if (DistributedWorker.debugWork) return console.warn("DistributedWorker.debugWork: Ignoring slice timeout");
      this.ee.emit('reject', new SliceTooSlowError(`Slice took longer than ${dcpConfig.worker.sliceTimeout / 1000} seconds.`));
    }, dcpConfig.worker.sliceTimeout);
  }
  /**
   * Handles progress and completion events from worker.
   * Unless explicitly returned out of this function will re-emit the event
   * on @this.ee where the name of the event is event.data.request.
   *
   * @param {object} event - event received from the worker
   */


  async onmessage(event) {
    if (DistributedWorker.debugEvents) {
      console.log('XXX eventDebug:', {
        id: this.id,
        request: event.data.request,
        state: this.state,
        event: event
      });
    }

    switch (event.data.request) {
      case 'debugLog':
        console.log("WORKER DEBUG LOG:", event.data);
        break;

      case 'dependency':
        let moduleData = await this.supervisor.fetchModule(event.data.data);
        this.worker.postMessage({
          request: 'moduleGroup',
          data: moduleData,
          id: event.data.id
        });
        break;

      case 'progress':
        let {
          progress,
          isIndeterminate
        } = event.data;
        this.progress = progress;
        this.time.push([progress, performance.now()]);
        this.resetProgressTimeout();
        this.emit('sliceProgress', {
          isIndeterminate,
          progress
        });
        break;

      case 'noProgress':
        let {
          message
        } = event.data;
        this.ee.emit('reject', new NoProgressError(message));
        break;

      case 'complete':
        clearTimeout(this.progressTimeout);
        clearTimeout(this.sliceTimeout);
        this.progressTimeout = this.sliceTimeout = null;

        if (this.progress === null) {
          // If a progress update was never received (progress === null) then reject
          this.ee.emit('reject', new NoProgressError('Worker never emitted a progress event.'));
          break;
        } else {
          this.progress = 100;
          this.time.push([100, performance.now()]);
        }

        event.data.time = this.time.slice(0);
        this.ee.emit('resolve', event);
        break;

      case 'uncaughtException':
        // console.log('393:XXX: uncaughtException, rejecting', event.data);
        this.emit('workerEmit', {
          eventName: 'uncaughtException',
          payload: event.data.error
        }); // Warning: rejecting here with just event.data.error causes issues
        // where the reject handlers modify the object so it interferes with the
        // workerEmit event payload, wrapping in an Error instance copies the values

        let e = new UncaughtExceptionError(event.data.error.message, event.data.error.fileName, event.data.error.lineNumber);
        e.stack = event.data.error.stack;
        e.name = event.data.error.name;

        if (this.ee.listenerCount('reject') > 0) {
          this.ee.emit('reject', e);
        } else {
          // This will happen if the error is thrown during initialization
          console.error("Unhandled worker exception:", e.message);
        }

        break;

      case 'accessListViolation':
        console.error("Worker access list violation:", event.data.message);
        break;

      case 'console':
        this.emit('workerEmit', {
          eventName: 'console',
          payload: event.data.payload
        });
        break;

      case 'emitEvent':
        /* ad-hoc event from the worker (supervisor.emit) */
        this.emit('workerEmit', {
          eventName: event.data.eventName,
          payload: event.data.payload
        });
        break;

      case 'checkCapabilitiesDone':
      case 'evalResult':
      case 'mainModuleEvaluated':
      case 'resetStateDone':
      case 'applyRequirementsDone':
        // emit internally
        this.ee.emit(event.data.request, event);
        break;

      case 'workerLoaded':
        // emit externally
        this.emit('workerLoaded', this);
        break;

      default:
        this.ee.emit('reject', new Error('Received unhandled request from worker: ' + event.data.request + '\n\t' + JSON.stringify(event)));
        break;
    }
  }
  /**
   * Error handler for the interal worker.
   * Currently just logs the errors that the worker spits out.
   */


  onerror(event) {} // console.error('391: Got a worker error! Boo!', event)

  /**
   * Clears the timeout and terminates the worker.
   *
   * @param {boolean} [reject = true] - if true call @reject
   */


  terminate(reject = true) {
    this.state = TERMINATED;
    clearTimeout(this.progressTimeout);
    clearTimeout(this.sliceTimeout);
    this.progressTimeout = this.sliceTimeout = null;

    if (this.worker && typeof this.worker.terminate === 'function') {
      try {
        this.worker.terminate();
        this.worker = null;
      } catch (e) {
        console.error("DistributedWorker.terminate: Encountered an error while terminating the worker:", e);
      } finally {
        this.emit('workerStop', this);
      }
    }

    if (reject) {
      this.ee.emit('reject', new Error(`Worker was terminated or timed out.`));
    }
  }
  /**
   * Attempts to stop the worker from doing completing its current
   * set of work without terminating the working.
   */


  stop() {
    throw new Error('DistributedWorker.stop is not yet implemented.');
  }
  /**
   * Runs a series of benchmarks inside of this worker.
   * Before running you must @start the worker.
   *
   * @returns {object} - an object with a series benchmark names -> time to complete
   */


  async benchmark() {
    let benchmarks = [];
    let timings = {
      total: 0
    };
    let benchmarkList = await protocol.justFetch(this.packageURL.resolve('benchmarks/LIST.json'), 'JSON');

    let flattenJSON = (obj, path = '') => {
      if (typeof obj === 'string') {
        return [path + obj];
      } else if (Array.isArray(obj)) {
        let paths = [];

        for (let file of obj) {
          paths.push(path + file);
        }

        return paths;
      } else {
        let paths = [];

        for (let dir in obj) {
          paths = paths.concat(flattenJSON(obj[dir], path + dir + '/'));
        }

        return paths;
      }
    };

    benchmarks = flattenJSON(benchmarkList);
    let name, data, test;

    while (benchmarks.length) {
      try {
        name = benchmarks.pop();
        data = '';

        if (name.startsWith('kraken')) {
          data = await protocol.justFetch(packageURL.resolve('benchmarks/${name}-data.js'));
        }

        test = await protocol.justFetch(packageURL.resolve('/benchmarks/${name}.js')); // the `if !print` addition is to get emscripten compiled benchmarks to work in the standalone miner

        let code = data + '\n' + test + `if (typeof print !== 'function') { print = console.log }`;
        let start = performance.now();
        await this.eval(code, `/benchmarks/${name}.js`);
        let runTime = performance.now() - start;
        let parts = name.split('/');
        let position = timings;

        for (let i = 0; i < parts.length - 1; i++) {
          if (!position[parts[i]]) {
            position[parts[i]] = {};
          }

          position = position[parts[i]];
        }

        position[parts[parts.length - 1]] = runTime;
        timings.total += runTime;
      } catch (error) {
        continue;
      }
    }

    return timings;
  }

}

exports.default = DistributedWorker;
DistributedWorker.idCounter = 1;

if (window && typeof __DCP_CLIENT__ == 'undefined') {
  window.DistributedWorker = DistributedWorker;
}
