"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

/** 
 * @file        event.js                An EventEmitter class which is API-compatible with Node's EventEmitter class.
 *
 * @author      Wes Garland, wes@page.ca
 * @date        Dec 2012, Jun 2018
 */
const LISTENERS = Symbol('EventEmitter Listeners');

class EventEmitter {
  constructor(debugLabel) {
    this.debugLabel = debugLabel || "";

    if (typeof require === 'function') {
      this.debug = require('./debugging').scope('event-emitter')(this.debugLabel || 'all');
    }

    this[LISTENERS] = {};
  }
  /** Add an event listener. 
   *  @param      eventName       The name of the event to capture, or true to capture all events
   *  @param      eventHandler    The function to invoke when the eventHandler is invoked.  The arguments
   *                              to the function are specified by the code emitting the event, however
   *                              a final argument, eventName, is appended to event handlers invoked via
   *                              catch-alls.        
   */


  addListener(eventName, eventHandler) {
    if (!this[LISTENERS].hasOwnProperty(eventName)) this[LISTENERS][eventName] = [];
    if (typeof eventHandler !== "function") throw new Error("eventHandler is not a function");
    this[LISTENERS][eventName].push(eventHandler);
    this.emit("newListener", eventName, eventHandler);
  }

  on(eventName, eventHandler) {
    this.addListener(eventName, eventHandler);
  }

  addEventListener(eventName, eventHandler) {
    this.addListener(eventName, eventHandler);
  }

  once(eventName, eventHandler) {
    function once_wrapper() {
      eventHandler.apply(eventHandler, arguments);
      this.removeListener(eventName, once_wrapper);
    }

    this.addListener(eventName, once_wrapper);
    return once_wrapper;
  }

  removeListener(eventName, eventHandler) {
    var i;
    if (!this[LISTENERS] || !this[LISTENERS].hasOwnProperty(eventName)) return;
    i = this[LISTENERS][eventName].indexOf(eventHandler);

    if (i != -1) {
      this[LISTENERS][eventName].splice(i, 1);
      if (this[LISTENERS][eventName].length === 0) delete this[LISTENERS][eventName];
    }
  }

  off(eventName, eventHandler) {
    this.removeListener(eventName, eventHandler);
  }

  removeEventListener(eventName, eventHandler) {
    this.removeListener(eventName, eventHandler);
  }

  removeAllListeners(eventName) {
    if (this[LISTENERS]) delete this[LISTENERS][eventName];
  }

  listeners(eventName) {
    if (!this[LISTENERS].hasOwnProperty(eventName)) this[LISTENERS][eventName] = [];
    return this[LISTENERS][eventName];
  }

  setMaxListeners(number) {
    throw new Error("not implemented");
  }

  listenerCount(eventName) {
    return this[LISTENERS][eventName] ? this[LISTENERS][eventName].length : 0;
  }
  /** Event emitter. All interfaces which emit events eventually fall through to this function. 
   *  This function invokes event listeners in the correct context for the passed eventName, as
   *  well as any catch-all (true) listeners.
   *
   *  @param      eventName       The name of the event to emit.
   *  @param      args            An array (not array-like) object containing the arguments 
   *                              to pass to the listener.
   */


  vemit(eventName, args) {
    var i;

    if (!this[LISTENERS] || !(this[LISTENERS].hasOwnProperty(eventName) || this[LISTENERS].hasOwnProperty(true))) {
      if (this.debug) console.log("Not firing " + this.debugLabel + "::" + eventName + " -- no listeners");
      return;
    }

    if (this.debug) console.log("Firing " + this.debugLabel + "::" + eventName);
    if (this[LISTENERS].hasOwnProperty(eventName)) for (i = 0; i < this[LISTENERS][eventName].length; i++) runSoon(this, this[LISTENERS][eventName][i], args);
    if (typeof this[LISTENERS][true] !== "undefined") for (i = 0; i < this[LISTENERS][true].length; i++) runSoon(this, this[LISTENERS][true][i], args.concat(eventName));
  }
  /** Emit an event. 
   *
   *  @param      eventName       The event to emit
   *  @param      ...             The arguments passed to the event handler
   */


  emit(eventName
  /* ... */
  ) {
    var args;
    args = Array.prototype.slice.call(arguments);
    args.shift();
    return this.vemit(eventName, args);
  }
  /** Proxy all of the events from another EventEmitter or DOM EventTarget and 
   *  re-emit them as though they originated from this EventEmitter.
   * 
   *  @param      element         The EventEmitter / EventTarget whose events we are to re-emit
   *  @param      eventName ...   One or more arguments which are strings that contain the names of
   *                              the events to re-emit.
   */


  proxy(element, eventName
  /*, eventName... */
  ) {
    var i, fun;

    var _this = this;

    for (i = 1; i < arguments.length; i++) {
      fun = function events$$EventEmitter$proxy$fun(thisEventName) {
        return function events$$EventEmitter$proxy$fun$bound() {
          if (typeof arguments[0] === 'object') arguments[0].proxyType = thisEventName;
          this.vemit(thisEventName, arguments);
        }.bind(_this);
      }(arguments[i]);

      element.addEventListener(arguments[i], fun);
    }
  }
  /** Proxy an events from another EventEmitter or DOM EventTarget and 
   *  re-emit them as though they originated from this EventEmitter, under a new event type.
   * 
   *  @param      element                 The EventEmitter / EventTarget whose events we are to re-emit
   *  @param      captureEventName        The name of the event to re-emit, or true to re-emit all events
   *  @param      newEventName            The event type of the new event when eventName is not true; when
   *                                      eventName is true, it is used as a prefix prepended to the real
   *                                      event name (ev.type)
   *
   *  *note* -    Catch-all proxies are only supported on instances of events$$EventEmitter. That
   *              functionality is not supported by DOM events.
   */


  proxyAs(element, captureEventName, emitterEventName) {
    var fun;
    if (element === this && captureEventName === true) throw new Error("Infinite recursion is not supported");

    fun = function events$$EventEmitter$proxyAs$eventHandler(ev) {
      var args = Array.prototype.slice.call(arguments);
      if (typeof ev === 'object' && ev.hasOwnProperty('type')) ev.proxyType = ev.type;
      if (captureEventName === true) this.vemit(emitterEventName + args.pop(), args);else this.vemit(emitterEventName, arguments);
    }.bind(this);

    element.addEventListener(captureEventName, fun);
  }
  /** Debug interface - not stable */


  dumpKnownEvents() {
    var eventName, i;
    var list = [];
    if (!this[LISTENERS] || !this[LISTENERS].length) return list;

    for (eventName in this[LISTENERS]) list.push(eventName);

    list.sort();

    for (i = 0; i < list.length; i++) {
      if (list[i] == list[i + 1]) list.splice(i + 1, 1);
    }

    return list;
  }

}
/** Run this supplied function ASAP in the reactor loop, with the provided 'this' and arguments.
 *  @param      _this   [optional]      An object to use apply as 'this' when invoking fn
 *  @param      fn                      The function to invoke
 *  @param      args                    Array of arguments passed to fn during invocation
 */


exports.default = EventEmitter;

function runSoon(_this, fn, args) {
  var callback;

  if (typeof _this === "function" && typeof args === 'undefined') {
    args = fn;
    fn = _this;
    _this = window;
  }

  if (typeof fn !== "function") throw new Error("fn is not a function");
  if (typeof args === 'object') callback = function runSoon_inner() {
    fn.apply(_this, args);
  };else callback = function runSoon_inner() {
    fn.apply(_this, []);
  };
  if (window) window.setTimeout(callback, 0);else setTimeout(callback, 0);
}
