"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.globalInstance = exports.default = exports.nsObject = void 0;

var _eventEmitter = _interopRequireDefault(require("./eventEmitter.js"));

var _keychain = _interopRequireDefault(require("./keychain.js"));

var _index = _interopRequireDefault(require("./wallet/index.js"));

var _bignumber = require("/node_modules/bignumber.js/bignumber.mjs");

var _messageToBuffer = require("./message-to-buffer.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 *  @file       protocol.js     Portable (browser, node) version of the DCPv3 Protocol.
 *                              
 *              This module exposes, as side-effects of loading, the following globals:
 *              - XMLHttpRequest        <= require('./dcp-xhr').XMLHttpRequest when node
 *              - io                    <= require('socket.io-client')
 *              - ethereumjs            <= require('ethereumjs-tx')
 *              - ethereumjs.Wallet     <= require('ethereumjs-wallet')
 *              - ethereumjs.Util       <= require('ethereumjs-util')
 *              - URL                   <= require('url') when node < 10(?)
 *              - protocol:             Shared instance of the Protocol class. This should be used for all DCP communications
 *              - Protocol:             main Protocol class, may be used to instantiate a secondary Protocol instance
 *              - BigNumber:            The DCP-standard BigNumber library; should be used for all math related to DCCs or money
 *
 *              Default export:         For environments where a default export works (future, maybe? modern 
 *                                      webpack seems not to do the right thing here), the CJS default export 
 *                                      will be the `protocol` shared instance
 *
 *              This module may also invoke fixups on the global dcpConfig when it is loaded; this implies 
 *              that dcpConfig must be loaded first, but not used until this module loads.
 *
 *  @author     Greg Agnew, gagnew@kingsds.network
 *              Matthew Palma, mpalma@kingsds.network
 *              Eddie Roosenmaallen, eddie@kingsds.network
 *  @date       Mar 2018
 *
 *  *note* -    This module contains significant browser-only code, such as modal dialog boxes, which
 *              currently do not behave reasonably on NodeJS.
 */

/* globals ethereumjs, XMLHttpRequest, io, dcpConfig, localStorage, performance, chrome, Modal */

/** Capture namespace pollution in dcp-client package without affecting global-happy legacy code (direct import or compute.min.js) */
let nsObject;
exports.nsObject = nsObject;

if (typeof __DCP_CLIENT__ !== 'undefined') {
  exports.nsObject = nsObject = {};
} else {
  exports.nsObject = nsObject = typeof window !== 'undefined' && window || typeof global !== 'undefined' && global || {};
}

if (typeof dcpConfig === 'undefined') {
  throw new Error('dcpConfig is not defined');
}

const DCPURL = require('./dcp-url');

const {
  requireNative
} = require('./webpack-native-bridge');
/* polyfill missing standard globals */


let _globalObject = typeof window !== 'undefined' && window || typeof global !== 'undefined' && global || nsObject;

if (typeof _globalObject.XMLHttpRequest === 'undefined') _globalObject.XMLHttpRequest = require('./dcp-xhr').XMLHttpRequest;
if (typeof _globalObject.URL === 'undefined') _globalObject.URL = require('url');

if (typeof nsObject.window === 'undefined') {
  nsObject.window = nsObject;
}

nsObject.ethereumjs = require('ethereumjs-tx');
nsObject.ethereumjs.Wallet = require('ethereumjs-wallet');
nsObject.ethereumjs.Util = require('ethereumjs-util');
nsObject.io = require('socket.io-client');
nsObject.BigNumber = _bignumber.BigNumber;

if (dcpConfig.needs && dcpConfig.needs.urlPatchup) {
  DCPURL.patchup(dcpConfig);
}

var publicMethods = ['/addresses', '/batch', '/getNonce'];
var version = '0.1.3';
const minVersion = '>=0.1';
var api = {
  replay: false,
  replayProtectedRoutes: [],
  acceptAddresses: [],
  useSockets: false,
  useExtension: false,
  batchRequests: true,
  version,
  minVersion
};
var nonces = {};
var origins = {};
var sockets = {};
var promises = {};
var messageBuffer = {};
var flushTimes = {};
var protocolEvents = {}; // count registered protocolEvent listeners, indexed by [origin][type] = nnn

const ID_SYMBOL = Symbol('Protocol Identity');
/**
 *  Format an error message that describes the current state of an error code originating
 *  from an XHR request.
 *
 *  @param      url             The URL associated with the request
 *  @param      xhr             The object passed to the onload / onerror / etc event handler
 *  @param      plain           True when we want to force non-colourized output
 *
 *  @returns    {string}        An error message, formatted with ANSI color when the output
 *                              is a terminal, suitable for writing directly to stdout. If
 *                              the response included html content (eg a 404 page), it is 
 *                              rendered to text in this string.
 */

function formatXHRErrorMessage(url, xhr, plain) {
  let chalk,
      message,
      headers = {};

  if (!plain && require('./dcp-env').platform === 'nodejs') {
    chalk = new requireNative('chalk').constructor({
      enabled: requireNative('tty').isatty(0)
    });
  } else if (!plain && require('./dcp-env').isBrowserPlatform) {
    chalk = {
      grey: a => `<font color="grey">${a}</font>`,
      yellow: a => `<font color="yellow">${a}</font>`,
      bold: a => `<bold>${a}</bold>`
    };
  } else {
    let nop = function nop(a) {
      return a;
    };

    chalk = {
      grey: nop,
      bold: nop,
      yellow: nop
    };
  }

  xhr.getAllResponseHeaders().replace(/\r/g, '').split('\n').forEach(function (line) {
    var colon = line.indexOf(': ');
    headers[line.slice(0, colon)] = line.slice(colon + 2);
  });
  message = `HTTP Status: ${xhr.status} for ${url}`;

  if (typeof headers['content-type'] !== 'undefined') {
    switch (headers['content-type'].replace(/;.*$/, '')) {
      case 'text/plain':
        message += '\n' + chalk.grey(xhr.responseText);
        break;

      case 'text/html':
        message += '\n' + chalk.grey(require('html-to-text').fromString(xhr.responseText, {
          wordwrap: parseInt(process.env.COLUMNS, 10) || 80,
          format: {
            heading: function (elem, fn, options) {
              var h = fn(elem.children, options);
              return '====\n' + chalk.yellow(chalk.bold(h.toUpperCase())) + '\n====';
            }
          }
        }));
        break;
    }
  }

  return message;
}
/**
 * The DCP Protocol; this class handles all communications between DCP agents
 * @module Protocol
 * @requires Constants
 */


class Protocol extends _eventEmitter.default {
  constructor() {
    super('Protocol');
    this.eth = nsObject.ethereumjs;
    this.keychain = new _keychain.default(this);
    this.connectedToExtension = null; // This is where the protocol identity is stored

    this[ID_SYMBOL] = {
      identity: null,
      // PK cached so that the user doesn't get
      // prompted for a passphrase with every sign
      privateKey: null
    };
    this.maxMessagesToFlush = 4;
    this.messageFlushDelayTime = 2000; // ms
  } // Return a list of all known origins


  get origins() {
    return origins;
  } // return the internal "version" string. May be associated with the "Protocol version",
  // but not currently directly related to the DCP Platform version


  get version() {
    return version;
  }
  /** Open a new websocket to an origin
   *  @param    origin          The origin to connect to
   *  @return   A Promise<boolean> which will resolve TRUE when the connection is complete,
   *                            or reject with the thrown error on connection error or timeout
   *  @todo: Reuse extant sockets, maybe?
   */


  connect(origin) {
    return new Promise((resolve, reject) => {
      if (sockets[origin]) return sockets[origin];
      let socket = nsObject.io(origin);
      socket.on('connect', () => {
        sockets[origin] = socket;
        this.sendOverSocket(new DCPURL.URL(origin).resolve('/addresses'), {}).then(signedResponse => {
          origins[origin] = signedResponse.message.api;
          resolve(true);
        });
      });
      socket.on('event', message => {
        this.emit(message.type, message.body);
      });
      socket.on('response', async message => {
        let results = message.result;

        if (!Array.isArray(results)) {
          results = [results];
        }

        for (let i = 0; i < results.length; i++) {
          try {
            await this.validate(results[i], {
              destination: false,
              // publicMethods.indexOf(message.url) !== -1,
              route: message.url
            });
          } catch (error) {
            // console.log(error)
            results.splice(i, 1, error);
          }
        }

        if (typeof promises[message.id] !== 'undefined') {
          promises[message.id].resolve(message.result);
          delete promises[message.id];
        }
      });
      socket.on('queued', message => {// log that transaction is queued
        // console.log(`request ${message.id} is queued`)
      });
      socket.on('error', message => {
        // console.log('server error', message)
        if (typeof promises[message.id] !== 'undefined') {
          promises[message.id].reject(message.body);
          delete promises[message.id];
        }
      });
      socket.on('connect_error', error => {
        console.error('320: connect error', error);
        this.emit('socketDisconnect', origin);
        reject(error);
      });
      socket.on('connect_timeout', error => {
        console.error('325: timeout error', error);
        this.emit('socketDisconnect', origin);
        reject(error);
      });
      socket.on('disconnect', reason => {
        // socket.io will automatically try to reconnect (if reasonable),
        // so don't delete the socket handle or we'll end up with many
        this.emit('socketDisconnect', origin, reason);
      }); // when we reconnect, re-subscribe to all registered protocolEvents

      socket.on('reconnect', () => {
        if (!protocolEvents[origin]) return;

        for (let type of Object.keys(protocolEvents[origin])) {
          this.listen(origin, type, {});
        }
      });
    });
  }
  /** Disconnect one or all websockets
   *  @param    origin          Optional origin to disconnect from. Default = disconnect all
   */


  disconnect(origin = false) {
    if (origin !== false) {
      if (typeof sockets[origin] !== 'undefined') {
        sockets[origin].disconnect(true);
        delete sockets[origin];
      }
    } else {
      let origins = Object.keys(sockets);

      for (let i = 0; i < origins.length; i++) {
        sockets[origins[i]].disconnect(true);
        delete sockets[origins[i]];
      }
    }
  }
  /** Fetch a new nonce for replay-protected server routes
   *  @param    url             URL to extract the origin from
   *  @param    privateKey      Private key to sign the request with. default = use active protocol key
   *  @return   Promise<string> Promise will resolve to the nonce, or reject with an error
   */


  async getNonceForOrigin(url, privateKey) {
    let options = new DCPURL.URL(url);
    let origin = `${options.protocol || window.location(window.location ? window.location.protocol : '')}//${options.hostname}`;

    if (options.port) {
      origin += ':' + options.port;
    }

    let response = await this.send(new DCPURL.URL(origin).resolve('/getNonce'), {}, privateKey, false, 0, false);
    return response.message.nonce;
  }
  /** Ask an origin for its signing address
   *  @param    url             URL from which to extract the origin
   *  @param    privateKey      Provide an optional private key to override the default signing key
   *  @return   Promise<string> Promise which will resolve to the origin's (first) address, or
   *                            reject with an error
   */


  async getAddressForOrigin(url, privateKey) {
    let options = new DCPURL.URL(url);
    let origin = `${options.protocol || window.location(window.location ? window.location.protocol : '')}//${options.hostname}`;

    if (options.port) {
      origin += ':' + options.port;
    }

    if (typeof origins[origin] !== 'undefined' && origins[origin].acceptAddresses.length > 0) {
      return origins[origin].acceptAddresses[0];
    } else {
      const t0 = Date.now(); // Keep a handle on the promise, so if we get more requests (for this origin) before the
      // first one is done, we still only send one request. Additional reqs will wait for the 
      // same promise, then all return the same data when it resolves.

      if (typeof origins[origin] === 'undefined' || !origins[origin].acceptAddressPromise) {
        origins[origin] = {
          acceptAddresses: []
        };
        origins[origin].acceptAddressPromise = this.send(new DCPURL.URL(origin).resolve('/addresses'), {}, privateKey, false, 0, false);
      } // console.log(`462: fetch addresses took ${Date.now() - t0}ms`);


      await origins[origin].acceptAddressPromise;
      return origins[origin].acceptAddresses[0];
    }
  }
  /** Subscribe to events from a DCP server
   *  @param    origin          Origin to listen on
   *  @param    type            Event name to subscribe to
   *  @param    message         Message to send with the listen request (default: {})
   *                            (@todo: what does this accomplish? how/is it used? ~ER20190624)
   *  @param    privateKey      Optionally override signing key
   *  @param    signed          Optional flag to indicate message is already signed.
   *  @return   Promise<boolean>  Promise will resolve to TRUE once listener is established, or 
   */


  async listen(origin, type, message = {}, privateKey = false, signed = false) {
    let finish = async () => {
      if (!protocolEvents[origin]) protocolEvents[origin] = {};
      if (!protocolEvents[origin][type]) protocolEvents[origin][type] = 0;
      let signedMessage = message;

      if (!signed) {
        if (typeof message.to === 'undefined') {
          message.to = await this.getAddressForOrigin(origin, privateKey);
        }

        signedMessage = await this.sign(message, privateKey);
      }

      return new Promise((resolve, reject) => {
        let messageId = this.eth.Util.bufferToHex(signedMessage.signature.r.data) + this.eth.Util.bufferToHex(signedMessage.signature.s.data) + signedMessage.signature.v;
        promises[messageId] = {
          resolve,
          reject
        };
        sockets[origin].emit('listen', {
          type,
          message: signedMessage
        });
        protocolEvents[origin][type]++;
        resolve(true);
      });
    };

    if (typeof sockets[origin] === 'undefined') {
      return this.connect(origin).then(success => {
        return finish();
      });
    } else {
      return finish();
    }
  }
  /** Unregister a protocol listener
   *  @param    origin          Origin to listen on
   *  @param    type            Event name to subscribe to
   *  Will decrement the reference counter for the origin/type, and will unlisten from the server
   *  when the count reaches zero
   */


  unlisten(origin, type) {
    if (!protocolEvents[origin]) return; // We are already not listening to this origin

    if (!protocolEvents[origin][type]) return; // We are already not listening to this event type on this origin

    protocolEvents[origin][type]--;

    if (protocolEvents[origin][type] <= 0) {
      sockets[origin].emit('unlisten', {
        type
      });
      delete protocolEvents[origin][type];
      if (Object.keys(protocolEvents[origin]).length === 0) delete protocolEvents[origin];
    }
  }

  async sendOverSocket(url, message, privateKey = false, signed = false, backOffTime) {
    let options = new DCPURL.URL(url);
    let origin = `${options.protocol || window.location(window.location ? window.location.protocol : '')}//${options.hostname}`;

    if (options.port) {
      origin += ':' + options.port;
    }

    if (publicMethods.indexOf(options.pathname) === -1) {
      let destination = await this.getAddressForOrigin(origin, privateKey);

      if (typeof message.to === 'undefined') {
        message.to = destination;
      }

      if (origins[origin].replay === true || origins[origin].replayProtectedRoutes.indexOf(options.pathname) !== -1) {
        if (typeof message.nonce === 'undefined') {
          message.nonce = await this.getNonceForOrigin(origin, privateKey);
        }
      }
    }

    let self = this;

    let finish = async function () {
      let signedMessage = message;

      if (!signed) {
        signedMessage = await self.sign(message, privateKey);
      }

      return new Promise((resolve, reject) => {
        let messageId = self.eth.Util.bufferToHex(signedMessage.signature.r.data) + self.eth.Util.bufferToHex(signedMessage.signature.s.data) + signedMessage.signature.v;
        promises[messageId] = {
          resolve,
          reject: error => {
            if (error.code && error.code === 'BADNONCE' && !signed && backOffTime <= 3200) {
              delete message.nonce;
              setTimeout(async () => {
                if (backOffTime === 0) backOffTime = 1;
                if (backOffTime === 1) backOffTime = 100;else backOffTime *= 2;
                this.sendOverSocket(url, message, privateKey, true, backOffTime).then(resolve, reject);
              }, backOffTime);
            } else {
              reject(error);
            }
          }
        };
        sockets[origin].emit('request', {
          route: options.pathname,
          message: signedMessage
        });
      });
    };

    if (typeof sockets[origin] === 'undefined') {
      return this.connect(origin).then(success => {
        return finish();
      });
    } else {
      return finish();
    }
  }

  addReplayProtection(...routes) {
    api.replayProtectedRoutes.push(...routes);
  }

  getNonce(address) {
    if (typeof nonces[address] === 'undefined') {
      nonces[address] = {
        current: 0,
        expected: {}
      };
    }

    let nonce = [nonces[address].current++, Math.random(), Date.now()].join('-');
    nonces[address].expected[nonce] = true;
    return nonce;
  }
  /**
   * Validates message come from the sender
   * @param {string} message
   */


  validate(signedMessage, options = {}) {
    return new Promise((resolve, reject) => {
      let publicKey = this.eth.Util.bufferToHex(this.eth.Util.ecrecover(this.eth.Util.hashPersonalMessage((0, _messageToBuffer.messageToBuffer)(JSON.stringify(signedMessage.message))), signedMessage.signature.v, signedMessage.signature.r.data, signedMessage.signature.s.data));
      let address = this.eth.Util.bufferToHex(this.eth.Util.pubToAddress(publicKey));

      if (address !== signedMessage.owner) {
        let error = new Error('Bad signature.');
        error.code = 'BADSIG';
        error.status = 401;
        return reject(error);
      }

      signedMessage.id = this.eth.Util.bufferToHex(signedMessage.signature.r.data) + this.eth.Util.bufferToHex(signedMessage.signature.s.data) + signedMessage.signature.v;

      if (options.destination) {
        if (typeof signedMessage.message.to === 'undefined') {
          let error = new Error('No destination.');
          error.code = 'NODEST';
          error.status = 401;
          return reject(error);
        }

        let destination = signedMessage.message.to.toLowerCase();

        if (destination[0] !== '0' && destination[1] !== 'x') {
          destination = '0x' + destination;
        }

        if (api.acceptAddresses.indexOf(signedMessage.message.to.toLowerCase()) === -1) {
          let error = new Error('Bad destination.');
          error.code = 'BADDEST';
          error.status = 401;
          return reject(error);
        }
      }

      if (api.replay === true && typeof options.route !== 'undefined' || api.replayProtectedRoutes.indexOf(options.route) !== -1) {
        if (typeof signedMessage.message.nonce === 'undefined' || typeof nonces[signedMessage.owner] === 'undefined' || nonces[signedMessage.owner].expected[signedMessage.message.nonce] === 'undefined') {
          let error = new Error('Bad nonce.');
          error.code = 'BADNONCE';
          error.status = 401;
          return reject(error);
        } else {
          delete nonces[signedMessage.owner].expected[signedMessage.message.nonce];
        }
      }

      resolve(true);
    });
  }
  /**
   * Addresses an object by signing with the scheduler key
   * @todo this assumes tasks are unique, are they? this will need nonce if they are not
   * @param {object} object
   * @param {string} [privateKey = false]
   */


  async address(object, privateKey = false) {
    let signedObject = await this.sign(object, privateKey);
    return '0x' + this.eth.Util.toBuffer(signedObject.signature.r.data).toString('hex') + this.eth.Util.toBuffer(signedObject.signature.s.data).toString('hex') + signedObject.signature.v.toString();
  }
  /**
   * Hashes and then signs a message
   * @param {string} body
   * @param {string} privateKey
   */


  async sign(body, privateKey = false) {
    // Here is the hierarchy of how the protocol identity is selected:
    // 1. If a privateKey was provided to sign, use it
    // 2. If keychain.currentAddress is set, use that
    // 3. If this protocol instance has an identity set use it. Otherwise,
    //     1. Use Wallet.get("id") to try and load id.keystore, if successful set the identity to that
    //     2. Generate a random temporary keystore and set it as the identity
    if (privateKey || this.keychain.currentAddress) {
      let ret = this.keychain.sign(body, privateKey);
      return ret;
    } else if (this[ID_SYMBOL].identity === null) {
      try {
        this[ID_SYMBOL].identity = await _index.default.getId();
      } catch (e) {
        console.warn("Failed to load identity.keystore, generating temporary protocol identity");
        this[ID_SYMBOL].identity = await new _index.default.Keystore(null, '');
      }
    }

    if (this[ID_SYMBOL].privateKey === null) {
      this[ID_SYMBOL].privateKey = (await this[ID_SYMBOL].identity.getPrivateKey()).valueOf();
    }

    return this.keychain.sign(body, this[ID_SYMBOL].privateKey);
  }
  /**
   * Adds a message to the Nagle (message) buffer if a batch request is being
   * made, else send the message normally.
   *
   * @param {string}  url                - The url to send the message to.
   * @param {*}       message            - The message to add to the buffer.
   * @param {string}  privateKey         - The message's signing key.
   * @param {boolean} signed             - True if the message is signed, false
   *                                       otherwise.
   * @param {number}  [flushBefore=1000] - The amount of time (ms) to wait
   *                                       before flushing the message buffer.
   *
   * @returns {Promise|*}                  Either a promise to add the message
   *                                       to the buffer, or the response to the
   *                                       message sent.
   */


  async addToBuffer(url, message, privateKey, signed, flushBefore = 1000) {
    // console.log('addToBuffer:681', url)
    if (!url.startsWith('http')) {
      // This should be removed
      url = `${dcpConfig.scheduler.location}${url}`;
    } // console.log('685> addToBuffer', url)


    let options;

    try {
      options = new DCPURL.URL(url); // console.log('addToBuffer:690:', options)
    } catch (error) {
      console.error('690:', error);
      console.error({
        url,
        message
      });
      throw error;
    }

    let origin = `${options.protocol || window.location(window.location ? window.location.protocol : '')}//${options.hostname}`;

    if (options.port) {
      origin += ':' + options.port;
    }

    if (typeof origins[origin] === 'undefined') {
      try {
        await this.getAddressForOrigin(origin, privateKey);
      } catch (error) {
        const err = new Error(`694: failed to get address (${error.message})`);
        err.originalError = error;
        throw err;
      }
    }

    let api = origins[origin];

    if (api.batchRequests !== true) {
      return this.send(url, message, privateKey, signed, false, 0, false);
    }

    if (typeof messageBuffer[origin] === 'undefined') messageBuffer[origin] = [];
    return new Promise((resolve, reject) => {
      // console.log('793+ adding to buffer', url)
      messageBuffer[origin].push({
        url,
        message,
        privateKey,
        signed,
        resolve,
        reject
      });
      if (flushBefore <= 0) this.flushBuffer();else this.flushBefore(flushBefore, origin);
    });
  }

  flushBefore(miliseconds, origin = false) {
    let _origin = origin || 'All';

    let timeToFlushBy = Date.now() + miliseconds;

    if (typeof flushTimes[_origin] === 'undefined') {
      flushTimes[_origin] = {
        willFlush: 0,
        timeout: null
      };
    }

    let times = flushTimes[_origin];

    if (times.timeout !== null) {
      if (times.willFlush < timeToFlushBy) return true;else clearTimeout(times.timeout);
    }

    times.willFlush = timeToFlushBy;
    times.timeout = setTimeout(this.flushBuffer.bind(this, origin), miliseconds);
  }
  /**
   * Flushes the buffer of messages from the specificed origin. If the origin is
   * not specified, flush all the message buffers.
   *
   * @param {boolean} [origin=false] - The origin of the messages in the buffer to flush.
   *
   * @returns {boolean}                True if there are no more messages inthe buffer, false otherwise.
   */


  async flushBuffer(origin = false) {
    // console.log('837> flushBuffer()');
    let origins = Object.keys(messageBuffer);

    for (let i = 0; i < origins.length; i++) {
      if (origin && origin !== origins[i]) continue;
      let to = await this.getAddressForOrigin(origins[i]);
      flushTimes[origin].timeout = null;
      if (typeof this.flushBuffer.timestamps === 'undefined') this.flushBuffer.timestamps = {};
      let messages = messageBuffer[origins[i]].splice(0);
      if (!messages || messages.length === 0) return true;
      messages.push({
        url: new DCPURL.URL(origin).resolve('/msg/check'),
        message: {
          timestamp: this.flushBuffer.timestamps[origin] || 0
        },
        resolve: response => {
          if (response.status === 'Timestamp OK') {
            switch (response.message.type) {
              case 'broadcast':
                if (response.message && typeof response.message.message !== 'undefined') {
                  this.emit('dcpmsg', response);
                }

                break;

              case 'command':
                this.emit('dcpmsg', response);
                break;

              case 'delete':
                this.emit('dcpmsg', response);
                break;
            }
          }
        },
        reject: function (response) {
          console.log('777: msg/check reject', response);
        }
      });
      this.flushBuffer.timestamps[origin] = Date.now();
      let batch = [];

      for (let j = 0; j < messages.length; j++) {
        if (j === this.maxMessagesToFlush) {
          messageBuffer[origins[i]].push(...messages.slice(j));
          let times = flushTimes[origins[i]];

          if (times.timeout === null) {
            times.timeout = setTimeout(this.flushBuffer.bind(this, origins[i]), this.messageFlushDelayTime);
            times.willFlush = Date.now() + this.messageFlushDelayTime;
          }

          break;
        }

        let batchMessage = messages[j];
        if (typeof batchMessage.message.to === 'undefined') batchMessage.message.to = to;
        batch.push({
          url: batchMessage.url,
          message: await this.sign(batchMessage.message, batchMessage.privateKey)
        });
      }

      const urls = batch.map(m => m.url.replace(/^http[s]?:\/\/[^/]+/, '')).join();
      const t0 = Date.now(); // console.log('897- sending /batch?' + urls)

      this.send(new DCPURL.URL(origin).resolve('/batch') + '?' + urls, batch, false, true, 0, false).then(response => {
        // console.log(`901= batch request took ${Date.now() - t0}ms`);
        if (!(response instanceof Array)) throw new Error('Server did not return an array of results.');

        for (let i = 0; i < response.length; i++) {
          if (typeof response[i].error !== 'undefined') messages[i].reject(response[i].error);else messages[i].resolve(response[i]);
        }
      }).catch(error => {
        for (let i = 0; i < messages.length; i++) {
          messages[i].reject(error);
        }
      });
    }
  }
  /**
   * Either sends a message to a url (signing the message if needed), or adds
   * it to the Nagle buffer (Lookup 'Nagle's algorithm' for the use of the
   * Nagle buffer).
   *
   * @param {string}  url                - URL to send message to
   * @param {object}  [message={}]       - Body of the message; default = {}
   * @param {string}  [privateKey=false] - Optionally override the signing key; default = use active protocol key
   * @param {boolean} [signed=false]     - TRUE if message is already signed; default = FALSE
   * @param {number}  [backOffTime=0]    - Optional delay before trying again
   * @param {boolean} [addToBuffer=true] - True if the message should be added to the Nagle buffer, false otherwise.
   *
   * @returns {*}                          A response to the sent message.
   */


  async send(url, message = {}, privateKey = false, signed = false, backOffTime = 0, addToBuffer = true) {
    // unless flagged to send the request immediately, use .addToBuffer
    if (addToBuffer) {
      return this.addToBuffer(url, message, privateKey, signed);
    }

    if (!url.startsWith('http')) {
      // This should be removed
      url = dcpConfig.scheduler.location.resolve(url);
    } // console.log('923> send', url);


    let options;

    try {
      options = new DCPURL.URL(url);
    } catch (error) {
      console.error('822:', error);
      console.error({
        url,
        message
      });
      throw error;
    }

    let origin = `${options.protocol || window.location(window.location ? window.location.protocol : '')}//${options.hostname}`;

    if (options.port) {
      origin += ':' + options.port;
    }

    if (publicMethods.indexOf(options.pathname) === -1) {
      let destination = await this.getAddressForOrigin(origin, privateKey);

      if (typeof message.to === 'undefined') {
        message.to = destination;
      }

      if (origins[origin].replay === true || origins[origin].replayProtectedRoutes.indexOf(options.pathname) !== -1) {
        if (typeof message.nonce === 'undefined') {
          const t_non = Date.now();
          message.nonce = await this.getNonceForOrigin(origin, privateKey); // console.log(`954: getNonceForOrigin took ${Date.now() - t_non}ms`)
        }
      }
    }

    let deeperErrorStack = new Error().stack;
    deeperErrorStack = deeperErrorStack.substring(deeperErrorStack.indexOf('\n') + 1);

    if (api.useSockets) {
      return this.sendOverSocket(url, message, privateKey, signed);
    } else {
      return new Promise(sendUsingXHR.bind(this));
    }
    /**
     * Sends the message using an XMLHttpRequest.
     * @param   resolve         Standard Promise callback
     * @param   reject          Standard Promise callback
     */


    function sendUsingXHR(resolve, reject) {
      let xhr = new XMLHttpRequest();
      xhr.open('POST', url);

      if (require('./dcp-env').platform === 'nodejs') {
        try {
          xhr.setRequestHeader('User-Agent', getUserAgent());
        } catch (e) {
          /* Some browsers will reject with "Refused to set unsafe header" */
        }
      }

      xhr.setRequestHeader('Content-type', 'application/json; charset=utf-8');
      /**
       * Handles the response to an XMLHttpRequest.
       *
       * @param {XMLHttpRequest} xhr - The XMLHttpRequest that has all of the
       *                               data requested in its response.
       */

      const onload = function handleResponseFromMessageSentUsingXHR(xhr) {
        xhr.removeEventListener('load', onload);
        xhr.removeEventListener('error', onerror); // Upon a successful response.

        if (xhr.status >= 200 && xhr.status < 300) {
          let result = JSON.parse(xhr.responseText);

          if (options.pathname === '/addresses') {
            origins[origin] = result.message.api;
          } // check hardcoded protocol version


          if (typeof result.message !== 'undefined' && typeof result.message.wrongVersion !== 'undefined' && result.message.wrongVersion && require('./dcp-env').isBrowserPlatform) {
            let lastRestart = window.localStorage.getItem('mining-restart');

            if (!lastRestart || Date.now() - lastRestart > 1000 * 60 * 5) {
              window.localStorage.setItem('mining-restart', Date.now());
              let prompt = new Modal('Version out of date', false, () => {
                window.location.reload(true);
              });
              prompt.addMessage('The browser will reload in 5 seconds.');
              prompt.open();
              setTimeout(() => {
                window.location.reload(true);
              }, 5000);
              return false;
            } else {
              let prompt = new Modal('Version out of date', false, () => {
                window.location.reload(true);
              });
              prompt.addMessage('Continue to reload browser to an updated version.');
              prompt.open();
              return false;
            }
          } // should validate the result here


          resolve(result);
        } else {
          // Upon a non-successful response on load, try to parse the
          // information in the response.
          let response;

          try {
            response = JSON.parse(xhr.responseText);
          } catch (jsonError) {
            let error = new Error(formatXHRErrorMessage(url, xhr));
            error.stack += '\n' + deeperErrorStack;
            error.status = xhr.status;
            return reject(error);
          } // Try to send the message again if we are able to.


          if (response.code === 'BADNONCE' && !signed && backOffTime <= 3200) {
            // try again
            delete message.nonce;
            setTimeout(async () => {
              if (backOffTime === 0) backOffTime = 1;
              if (backOffTime === 1) backOffTime = 100;else backOffTime *= 2; // not sure what will happen to this if unhandled.. should bubble up?

              this.send(url, message, privateKey, false, backOffTime).then(resolve, reject);
            }, backOffTime);
          } else {
            // When the request was unsuccessful and we can't send the message
            // again.
            const error = new Error(formatXHRErrorMessage(url, xhr));
            error.status = xhr.status;
            error.code = response.code;
            error.stack += '\n' + deeperErrorStack;

            if (response.error) {
              error.error = response.error;

              if (response.error.stack) {
                error.remoteStack = response.error.stack;
              }
            }

            reject(error);
          }
        }
      }.bind(this, xhr);

      let onerror = function (xhr) {
        let error = new Error(formatXHRErrorMessage(url, xhr));
        error.stack += '\n' + deeperErrorStack;
        error.status = xhr.status;
        xhr.removeEventListener('load', onload);
        xhr.removeEventListener('error', onerror);
        reject(error);
      }.bind(this, xhr);

      xhr.addEventListener('load', onload);
      xhr.addEventListener('error', onerror);

      if (signed) {
        // console.log('1115: already signed, no overhead')
        xhr.send(JSON.stringify(message));
      } else {
        // const t_crypto = Date.now();
        this.sign(message, privateKey).then(signedMessage => {
          // console.log(`1121: signing took ${Date.now() - t_crypto}`);
          // console.log('is valid:', this.validate(signedMessage), signedMessage)
          xhr.send(JSON.stringify(signedMessage));
        }, reject);
      }
    }
  }
  /**
   * Unlocks a password protected keystore file
   * @param {string|object} keystore - An Ethereum V3 keystore file
   * @param {string} password
   * @returns {object} - The unlocked wallet
   */


  unlock(keystore, password) {
    if (typeof keystore === 'object') keystore = JSON.stringify(keystore);
    return this.eth.Wallet.fromV3(keystore, password, true);
  }
  /**
   * Creates a new wallet object from privatekey or generates a new one
   * @param {string} [privateKey = false]
   */


  createWallet(privatekey = false) {
    if (privatekey) {
      return this.eth.Wallet.fromPrivateKey(this.eth.Util.toBuffer(privatekey));
    } else {
      return this.eth.Wallet.generate();
    }
  }
  /**
   * Will create a string keystore file from a wallet and password
   * @param {object} wallet
   * @param {string} password
   */


  createKeystore(wallet, password) {
    return wallet.toV3(password, {
      n: 1024
    });
  }

  setApi(_api) {
    let keys = Object.keys(_api);

    for (let i = 0; i < keys.length; i++) {
      api[keys[i]] = _api[keys[i]];
    }

    if (_api.privateKey) {
      this.keychain.addPrivateKey(_api.privateKey, true);
    }

    for (let i = 0; i < api.acceptAddresses.length; i++) {
      api.acceptAddresses[i] = api.acceptAddresses[i].toLowerCase();
    }
  }

  getApi(unsafe = false) {
    if (unsafe) {
      return api;
    } else {
      return {
        acceptAddresses: api.acceptAddresses,
        replay: api.replay,
        replayProtectedRoutes: api.replayProtectedRoutes,
        useSockets: api.useSockets,
        version: api.version,
        minVersion: api.minVersion,
        batchRequests: api.batchRequests
      };
    }
  }
  /**
   * Make a request, as a promise, resolves with a string or json object
   *
   * @param {string} url - path of file to be fetched
   * @param {string} [type = "string"] - expected type of the response
   * @param {string} [method = "GET"] - type of request to make
   * @param {boolean} [bustCache = false] - whether to add cache poison to the end of the url
   * @param {any} [body] - what to send as the body of the message
   * @returns {Promise} - resolves with response (json if type="JSON", string otherwise), rejects on fail
   */


  justFetch(url, type = 'string', method = 'GET', bustCache = false, body = undefined) {
    return new Promise((resolve, reject) => {
      let deeperErrorStack = new Error().stack;
      deeperErrorStack = deeperErrorStack.substring(deeperErrorStack.indexOf('\n') + 1);
      if (typeof url === 'object' && url.href) url = url.href;

      if (bustCache) {
        url += '?' + (typeof document !== 'undefined' && document.head && document.head.getAttribute('version') || Date.now());
        /* cache buster */
      }

      let xhr = new XMLHttpRequest();

      xhr.onloadend = function Protocol$$justFetch$onloadend() {
        try {
          delete xhr.onloadend;

          if (xhr.status >= 200 && xhr.status < 300) {
            let data = xhr.responseText;

            if (type === 'JSON') {
              data = JSON.parse(data);
            }

            resolve(data);
          } else {
            let error = new Error(`HTTP Error ${xhr.status} fetching ${url}`);
            error.request = xhr;
            error.request.location = url;
            error.request.method = method;
            error.stack += '\n' + deeperErrorStack;
            throw error;
          }
        } catch (e) {
          reject(e);
        }
      };

      xhr.open(method, url);

      if (typeof body !== 'undefined') {
        xhr.open(method, url);
        xhr.setRequestHeader('Content-type', 'application/json; charset=utf-8');
        xhr.send(JSON.stringify(body));
      } else {
        xhr.send();
      }
    });
  }

}

exports.default = Protocol;

function getUserAgent() {
  const path = require('path');

  const argv = process ? process.argv.filter(p => path.basename(p) !== 'ProcessContainerFork.js') : [];
  var uaTag = false;

  if (typeof navigator !== 'undefined' && navigator.userAgent) {
    uaTag = navigator.userAgent;
  } else if (argv[1]) {
    uaTag = path.basename(argv[1]);
  } else if (argv[0]) {
    uaTag = path.basename(argv[0]);
  }

  if (uaTag) {
    return `DCP Protocol; (${uaTag})`;
  }

  return 'DCP Protocol (general)';
}

const globalInstance = new Protocol();
exports.globalInstance = globalInstance;
nsObject.protocol = globalInstance;
nsObject.Protocol = Protocol;

if (typeof module !== 'undefined') {
  let descriptor = Object.getOwnPropertyDescriptor(module, 'exports');
  if (typeof descriptor !== 'undefined' && descriptor.writable) module.exports = globalInstance;
}
