/**
 *  @file       eth.js      
 *              Library code for manipulating ethereum keys, addresses, etc
 *              with syntactic sugar and abstractions for common but error-prone
 *              operations like equivalence comparison.
 *
 *  @module     eth
 *  @author     Wes Garland, wes@kingds.network
 *  @date       July 2019
 */
"use strict"
const utils = require('ethereumjs-util');
const isBuffer = require('is-buffer');  /* Required due to deficiencies in webpack implementation. /wg,rr oct 2019 */

/**
 *  Class which represents an Ethereum address (public key); used in DCP for public
 *  mentions of things like proof of user identity, bank account numbers, etc.
 *
 *  @constructor
 *  @param address      {string|object} The address we want to represent. If the
 *                                      passed address is a string, it should be
 *                                      (but is not required to be) in checksum format;
 *                                      a leading 0x is not required.  If the address
 *                                      is an object, we coerce the object with 
 *                                      toString(16) before working with it (BigNumber
 *                                      and similar objects are expected to "just work")
 *
 * *note* -     An address which arrives as an all-uppercase string and no leading 0x
 *              will not have its checksum validated.  This is to ease debugging against
 *              MySQL database copy-and-paste and should have a negligible real-world
 *              impact.
 * @param       testOnly        when true, return true only when the privateKey argument could 
 *                              be used to construct without throwing
 *
 * @throws      TypeError       when the address is invalid, fails checksum, etc.
 * @throws      TypeError       when the address is a number; eg a programmer typed in
 *                              a hexadecimal literal.
 * @returns     instance of Address whose value is a checksummed string of hex
 *              digits with leading 0x.
 */
exports.Address = function dcpEth$$Address(address, testOnly) {
  let cksumAddress
  
  if (typeof address === 'undefined' || address === null) {
    if (testOnly)
      return false
    throw new TypeError('Invalid address')
  }
  
  if (typeof address === 'number') {
    if (testOnly)
      return false
    throw new TypeError('There is not enough resolution in a Number to store an Ethereum address')
  }
  
  if (typeof address !== 'string') {
    if (typeof Buffer !== 'undefined' && isBuffer(address))
      address = address.toString('hex')
    else
      address = address.toString(16)
  }

  if (address.match(/^0x/))
    address = address.slice(2)
  
  if (!address.match(/^[a-fA-F0-9]*$/)) {
    if (testOnly)
      return false
    throw new TypeError(`Address must be composed of only hex digits; ${address} is invalid`)
  }
  
  if (address.length !== 40) {
    if (testOnly)
      return false
    throw new TypeError(`Addresses must be 40 digits long; ${address} is ${address.length} digits`)
  }
  
  address = '0x' + address.padStart(40, 0)
  
  if (address.toLowerCase() === address) {
    cksumAddress = utils.toChecksumAddress(address)
  } else {
    // Prevent 0X prefix from always failing the uppercase check.
    if (address === `0x${(address.slice(2)).toUpperCase()}`) {
      cksumAddress = utils.toChecksumAddress(address)
    } else {
      if (address !== utils.toChecksumAddress(address)) {
        if (testOnly)
          return false
        throw new TypeError(`Bad checksum for address ${address}`)
      }
      cksumAddress = address
    }
    address = address.toLowerCase()
  }

  if (testOnly)
    return true

  Object.defineProperty(this, 'valueOf', {
    value: function dcpEth$$Address$valueOf() {
      return cksumAddress
    },
    configurable: false,
    writable: false,
    enumerable: false
  })
}

exports.isAddress = function dcpEth$$isAddress(address) {
  return exports.Address(address, true)
}
  
/**
 *  Equality comparsion.
 *  @param address2     {string|object} The address we want to compare. May be
 *                                      used as an argument to the Address()
 *                                      constructor.
 *  @param      address2        {string|object} 
 *  @returns    true if address2 represents the same eth address as this instance;
 *              false otherwise.
 */
exports.Address.prototype.eq = function dcpEth$$Address$eq(address2) {
  let address = this.valueOf()
  if (this === address2 || address === address2)
    return true
  address2 = new (exports.Address)(address2).toString()
  return this.valueOf() === new (exports.Address)(address2).valueOf()
}

/** toString() method.
 *  @returns An Ethereum address in checksum format, with a leading 0x.
 *
 *  The toJSON alias is to ensure that Address instances will be coerced
 *  to a string by JSON.stringify.
 */
exports.Address.prototype.toJSON =
exports.Address.prototype.toString = function dcpEth$$Address$toString() {
  return this.valueOf()
}

/** Corresponds-To method.
 *  @param      privateKey      An instance of PrivateKey or a value suitable for
 *                              the PrivateKey constructor
 *  @returns    true when this address corresponds to the passed privateKey
 */
exports.Address.prototype.ct = function dcpEth$$Address$ct(privateKey) {
  return this.eq(new (exports.PrivateKey)(privateKey).toAddress())
}

/**
 *  Class which represents an Ethereum private key used in DCP for things
 *  like proof of user identity, bank account ownership, etc.
 *
 *  @constructor
 *  @param privateKey   {string|object} The private key we want to represent. If the
 *                                      passed private key is a string, it should (but
 *                                      is not required to) have the 0x prefix. If 
 *                                      privateKey is an object, we coerce the object with 
 *                                      toString(16) before working with it (BigNumber
 *                                      and similar objects are expected to "just work")
 * @param       testOnly        when true, return true only when the privateKey argument could 
 *                              be used to construct without throwing
 * @throws      TypeError       when privateKey is invalid
 * @throws      TypeError       when privateKey is a number; eg a programmer typed in
 *                              a hexadecimal literal.
 * @returns     instance of PrivateKey whose value is a checksummed string of hex
 *              digits with leading 0x.
 */
exports.PrivateKey = function dcpEth$$PrivateKey(privateKey, testOnly) {

  if (typeof privateKey === 'undefined' || privateKey === null) {
    if (testOnly)
      return false
    throw new TypeError('Invalid private key')
  }

  if (typeof privateKey === 'number') {
    if (testOnly)
      return false
    throw new TypeError('There is not enough resolution in a Number to store an Ethereum private key')
  }
  
  if (typeof privateKey !== 'string') {
    if (typeof Buffer !== 'undefined' && isBuffer(privateKey))
      privateKey = privateKey.toString('hex')
    else
      privateKey = privateKey.toString(16)
  }

  if (privateKey.match(/^0x/))
    privateKey = privateKey.slice(2)
  
  if (!privateKey.match(/^[a-fA-F0-9]*$/)) {
    if (testOnly)
      return false
    throw new TypeError(`Private key must be composed of only hex digits; ${privateKey} is invalid`)
  }

  if (privateKey.length !== 64) {
    if (testOnly)
      return false;
    throw new TypeError(`Private key ${privateKey} is too short`);
  }
  
  if (testOnly)
    return true

  privateKey = '0x' + privateKey.padStart(64, 0).toLowerCase()
  try {
    Object.defineProperty(this, 'valueOf', {
      value: function dcpEth$$PrivateKey$valueOf() {
        return privateKey
      },
      configurable: false,
      writable: false,
      enumerable: false
    }) } catch(e) {
    throw new Error('PrivateKey constructor must be invoked with new operator')
  }
}

exports.isPrivateKey = function dcpEth$$isPrivateKey(pk) {
  return exports.PrivateKey(pk, true)
}
  
/**
 *  Equality comparsion.
 *  @param privateKey2     {string|object}    The private key we want to compare. May be
 *                                            used as an argument to the PrivateKey()
 *                                            constructor.
 *  @returns    true if privateKey2 represents the same eth private key as this instance;
 *              false otherwise.
 */
exports.PrivateKey.prototype.eq = function dcpEth$$PrivateKey$eq(privateKey2) {
  let privateKey = this.valueOf()
  if (this === privateKey2 || privateKey === privateKey2)
    return true
  return privateKey.toString() === new (exports.PrivateKey)(privateKey2).toString()
}

/** Corresponds-To method.
 *  @param      address         An instance of Address or a value suitable for the 
 *                              Address constructor
 *  @returns    true when this address corresponds to the passed privateKey
 */
exports.PrivateKey.prototype.ct = function dcpEth$$PrivateKey$ct(address) {
  return this.toAddress().eq(address)
}

/** toString() method.
 *  @returns An Ethereum private key including leading 0x.
 */
exports.PrivateKey.prototype.toString = function dcpEth$$PrivateKey$toString() {
  return this.valueOf()
}

/** 
 * Determine the address (public key) associated with this instance of PrivateKey.
 * Only performs the hash once, no matter how many times you call the method.
 * *note* - This has a side effect of converting the method call from an inherited
 *          property to an own property on first call.
 *
 * @returns instance of Address
 */
exports.PrivateKey.prototype.toAddress = function PrivateKey() {
  let address = new (exports.Address)(utils.privateToAddress(this.toString()))
  Object.defineProperty(this, 'toAddress', {
    value: function dcpEth$$PrivateKey$toAddress() {
      return address
    },
    configurable: false,
    writable: false,
    enumerable: false
  })
  return this.toAddress()
}

const inspect = Symbol.for('nodejs.util.inspect.custom');
exports.PrivateKey.prototype[inspect] = function () { return '[Object dcpEth$$PrivateKey(' + this.toString() + ')]' }
exports.Address.prototype[inspect] = function () { return '[Object dcpEth$$Address(' + this.toString() + ')]' }
