/**
 *  @file       wallet.js
 *              Wallet API - perform operations related to Addresses, Keystores, Key Pairs
 * 
 *              DCP-741: https://kingsds.atlassian.net/browse/DCP-741
 *
 *  @author     Duncan Mays    - duncan@kingsds.network
 *  @author     Badrdine Sabhi - badr@kingsds.network
 *  @date       August 2019
 */

const _ConstructOnly = Symbol('Construct Only')
const isBuffer = require('is-buffer');  /* Required due to deficiencies in webpack implementation. /wg,rr oct 2019 */
const debugging = require('../debugging').scope('wallet');

if (require('../dcp-env').platform === 'nodejs') { 
  const { requireNative, } = require('../webpack-native-bridge')

  var path = requireNative('path')
  var fs = requireNative('fs')
  var os = requireNative('os')
  var pprompt = requireNative('password-prompt')
  var expandTilde = requireNative('expand-tilde')

  debug = !!(requireNative('process').env.DCP_DEBUG || '').match(/wallet|\*|verbose/)
} else if (require('../dcp-env').isBrowserPlatform) {
  debug = !!(document.body.getAttribute("dcp-debug") || '').match(/wallet|\*|verbose/)
}

Object.assign(exports, require('./eth'))
Object.assign(exports, require('./address-collection'));

/**
 *  This property sets the name of the namespace that is used to generate the application data directory, which is used to store proxy keys when required,
 *  and users’ bank and identity keys when supplied.
 *  The default namespace name is the name specified in the package.json file for the application calling into this API.
 */
exports.namespaceName = undefined

/**
 *  This property gets/sets the full path name the application data directory, overriding namespaceName if it is changed.
 *  The default value is ~/.dcp/applications/${namespaceName}/, where ~ represents the invoking user’s home directory.
 */
exports.stateDir = undefined

/**
 *  This function accepts a Keystore filename or name + options objects, reads the associated 
 *  keystore file, and returns and object
 *
 *  @returns  Promise which resolves to an object, which has the following properties:
 *                                      keystore - a Keystore Object corresponding to the keystore that was loaded.
 *                                      safe - a boolean flag which, when true, indicates that the keystore was read from a secure location.
 */
exports.load = async function wallet$$load() {
  let keystore, label
  let filename;
  let KeystoreConstructor = exports.Keystore

  if (typeof arguments[0] === 'string') {
    filename = exports.expandFilePathSymbols(arguments[0]);
    label = path.basename(filename, '.keystore')
  } else {
    /* form 2: build filename */
    let defaultOptions = {
      name: "default",
      paths: [ dcpConfig.global.hasOwnProperty('dotDcpDir')
               ? exports.expandFilePathSymbols(dcpConfig.global.dotDcpDir)
               : path.join(os.homedir(), '.dcp')
             ]
    }

    let options = arguments[0] || {}
    if (options.dir)
      options.paths = [options.dir]

    Object.assign(defaultOptions, options)
    options = defaultOptions

    if (exports.stateDir)
      options.paths.unshift(path.resolve(exports.stateDir))
    else if (exports.namespaceName)
      options.paths.unshift(path.resolve(os.homedir(), '.dcp', 'applications', exports.namespaceName))

    for (let i=0; i < options.paths.length; i++) {
      let fn = path.resolve(path.join(options.paths[i], options.name + '.keystore'));
      debugging('search') && console.log(`wallet - checking ${fn}`);
      if (fs.existsSync(fn)) {
        filename = fn
        debugging('searchResult') && console.log(`wallet - found keystore at ${fn}`);
        break
      }
    }

    if (!filename) {
      let err = new Error('Could not locate ' + options.name + '.keystore')
      debugging('searchResult') && console.log(`wallet - could not find keystore ${options.name}.keystore`);
      err.code = 'ENOENT'
      throw err;
    }

    label = options.name
    if (options.hasOwnProperty('KeystoreConstructor'))
      KeystoreConstructor = options.KeystoreConstructor
  }

  if (!filename) {
    throw new Error('filename not specified')
  }

  debugging() && console.log(`wallet - loading '${filename}'`);
  //gets the keystore or private key from the file
  keystore = await new KeystoreConstructor(fs.readFileSync(filename, 'utf-8').trim(), false);
  if (!keystore.label)
    keystore.label = label
  
  //return keystore in an object with a safety indicator
  return {
    keystore: keystore,
    safe: exports.isSafe(filename)
  }
}

/**
 *  @param      keystore {Object}       instance of Keystore to save
 *  @param      filename {String}       the filename to save it in
 *  @param      options  {Object}       optional options object to pass to node's
 *                                      file.writeFile().  Note that the mode will
 *                                      always force utf-8 output.
 */
exports.save = function wallet$$save(keystore, filename, mkdirOptions, writeFileOptions) {
  return new Promise(function(resolve, reject) {
    try {
      if (mkdirOptions !== false) {
        if (!mkdirOptions)
          mkdirOptions = { recursive: true }
        if (!fs.existsSync)
          fs.mkdirSync(mkdirOptions)
      }

      writeFileOptions = Object.assign({/*defaults*/}, writeFileOptions)
      writeFileOptions.encoding = 'utf-8'
      
      function cb(err) {
        if (err)
          reject(err)
        else
          resolve()
      }
      
      fs.writeFile(filename, JSON.stringify(keystore), writeFileOptions, cb)
    } catch(e) {
      reject(e)
    }
  })
}

/**
 * This function checks the safety requirements outlined in the Keystore API documentation
 * The requirements are that the file must neither be world readable or world writable, 
 * or have any parent directory that is world writable. This function will return a boolean which indicates
 * weather or not these conditions are met.
 *
 * @param   {string}  - filePath, this is the path to the file we are checking the safety of, this
 *                      function will not work properly, and in fact will log a warning message, if that filePath 
 *                      starts with either . or ..
 * @returns {boolean} - indicates weather or not the file provided satisfies safety criteria
 */
exports.isSafe = wallet$$isSafe = filePath => {
  if (!path.isAbsolute(filePath)) {
    throw new Error(`Not an absolute path: '${filePath}'`);
  }

  let pathArray = filePath.split(path.sep)
  pathArray[0] = pathArray[0] || path.parse(filePath).root;
  if (exports.isWorldReadable(filePath))
    return false

  //checks that the file and all its ancestor directories are not world readable
  while (pathArray.length > 0) {
    let reassemble = path.join.apply(null, pathArray)
    if (exports.isWorldWritable(reassemble))
      return false
    pathArray = pathArray.slice(0, pathArray.length - 1)
  }
  return true
}

/**
 * This function takes a filepath and returns a boolean to indicate weather or not the location specified in the filepath is world readable
 *
 * @param   {string}  filePath  - path to the file in question, does not work right if the filePath starts with . or ..
 *
 * @returns {boolean}           - indicates weather the given file or directory is world readable
 */
exports.isWorldReadable = wallet$$isWorldReadable = filePath => {
  const mode = fs.statSync(filePath).mode
  return mode & 4
}

/**
 * This function takes a filepath and returns a boolean to indicate weather or not the location specified in the filepath is world writable
 *
 * @param   {string}  filePath - path to the file in question, does not work right if the filePath starts with . or ..
 *
 * @returns {boolean}          - indicates weather the given file or directory is world writable.
 */
exports.isWorldWritable = wallet$$isWorldWritable = filePath => {
  debugging('safety') && console.debug(`wallet - checking write permissions on '${filePath}'`);
  const mode = fs.statSync(filePath).mode
  return mode & fs.constants.W_OK;
}

/**
 * This function takes a string, prompt, prints it to the screen, and then returns whatever the user types after it.
 * @param   {string} prompt - what gets displayed to the user.
 * @returns {string}        - what the user types.
 */
wallet$$passphrasePrompt_browser = (label, urgent, skipCheckFn) => new Promise((resolve, reject) => {
  let promptMessage = `Enter passphrase to unlock keystore '${label}': `
  const prompt = new (require('../modal')).default('Unlock Keystore', promptMessage, onContinue, handleExitingUpload)

  prompt.addFormElement({
    placeholder: 'Passphrase',
    type: 'password',
    autocomplete: 'on'
  });
  prompt.open()

  function onContinue(formValues) {
    const [password] = formValues;
    resolve(password);
  }

  function handleExitingUpload() {
    reject(new Error("Keystore unlock was cancelled."));
  }
})

function ppServiceList() {
  let head = ppServiceList.pending[0]

  if (ppServiceList.pending.length === 0) {
    clearTimeout(ppServiceList.timer)
    ppServiceList.timer = undefined
    return
  }
  
  if (head.pending) {
    if (!ppServiceList.timer)
      ppServiceList.timer = setTimeout(ppServiceList, 100)
    return
  }
  
  head.pending = true
  function afterPrompt(ppPromise) {
    ppServiceList.timer = setTimeout(ppServiceList, 100)
    let idx = ppServiceList.pending.indexOf(head)
    if (idx !== -1)
      ppServiceList.pending.splice(idx, 1)
    head.resolve(ppPromise)
  }

  if (head.skipCheckFn && head.skipCheckFn())
    return afterPrompt()

  pprompt(head.prompt, { method: 'hide', required: false }).then(afterPrompt)
}
ppServiceList.pending = []
ppServiceList.timer

function wallet$$passphrasePrompt_node(label, urgent, skipCheckFn) {
  return new Promise(function(resolve, reject) {
    try {
      /* force the outputting of pprompt() onto the event loop, so that pending 
       * console.log messages (etc) get printed first, when the loop is serviced,
       * instead of while we are typing the password.
       *
       * Also, make it so that it stalls on the event loop if there is already a
       * a password prompt being displayed.
       */
      let listO   = {
        resolve: resolve,
        reject: reject,
        prompt: `Enter passphrase to unlock keystore '${label}': `,
        skipCheckFn: skipCheckFn
      }

      if (ppServiceList.pending.length === 0 || urgent)
        setTimeout(ppServiceList, 0)
      if (urgent)
        ppServiceList.pending.unshift(listO)
      else
        ppServiceList.pending.push(listO)
    } catch(e) {
      reject(e)
    }
  })
}

exports.passphrasePrompt = function wallet$$passphrasePrompt() {
  debugging('stack') && console.debug('wallet - passphrasePrompt call site ', new Error('debug\b\b\b\b\b\b\b\b\b\b\b\b            '));
  if (require('../dcp-env').platform === 'nodejs')
    return wallet$$passphrasePrompt_node.apply(this, arguments);
  else
    return wallet$$passphrasePrompt_browser.apply(this, arguments);
}



/**
 *  This function takes a file path with symbols like . .. and ~, and expands them into the
 *  propper paths that they represent, so ~ turns into the user's home directory, . into 
 *  the current directory, and so on.
 *
 *  @param   {string} filePath
 *
 *  @returns {string} the expanded filepath
 */
exports.expandFilePathSymbols = wallet$$expandFilePathSymbols = filePath => {
  //Expands a leading tilde in a file path to the user home.
  filePath = expandTilde(filePath)
  //Takes out any . or ..s
  return path.resolve(filePath)
}

/**
 * These functions examine parameters and accepts options relating to its environment, uses this information 
 * to select a Keystore File, load it via wallet.load(), and return an AuthKeystore object.
 * If it cannot return an AuthKeystore object, this function will throw an Error.
 *
 * @param   {object | string} param1 | param2 - Three accepted keys: name, which specifies the name of the keystore, dir, which specifies its directpry, 
 *                                              and checkEmpty, which controls weather or not unlock() will attempt an empty password on the keystore 
 *                                              before prompting the user.
 * @returns {Object}                          - AuthKeystore object
 */

wallet$$get_browser = function (argv, options) {
  let KeystoreConstructor = options.KeystoreConstructor
  
  return new Promise((resolve, reject) => {

    if (require('../dcp-env').isBrowserPlatform && KeystoreConstructor === exports.IdKeystore) {
      /** XXX @todo implement identity keys on web platforms  /wg oct 2019 */
      resolve(new exports.IdKeystore(null, ''))
      return
    }

    if (window.sessionStorage.getItem('keystore')) {
      resolve(new exports.AuthKeystore(window.sessionStorage.getItem('keystore'), window.sessionStorage.getItem('passphrase') || ''))
      return
    }
    
    const promptMessage = `This file corresponds to your bank account on the Distributed Computer. You can download ` +
                          `your keystore file from the portal at ${dcpConfig.portal.location.href} via the Wallet page.`;
    const prompt = new (require('../modal')).default('Upload Keystore', promptMessage, onContinue.bind(this), handleExitingUpload)

    let keystore
    prompt.addFormElement({
      label: 'Choose a wallet file to upload (e.g. .keystore, .json):',
      type: 'file',
      callback: (file) => {
        // Read the file selected by the file input whenever the selection
        // changes.
        let reader = new FileReader()
        reader.onload = () => { keystore = reader.result }
        reader.readAsText(file)
      }
    }, {
      placeholder: 'Passphrase',
      type: 'password'
    },/*{
      type: 'label',
      label: 'Or enter a private key:'
    },*/ {
      type: 'text',
      placeholder: '0x...',
      pattern: '^(0x)?([a-fA-F0-9]){64}$',
      minLength: 64,
      maxLength: 66
    })
    // To make the user fill in at least one of the forms.
    prompt.changeFormSubmitHandler(validateOnSubmit)
    // parent.parent to hide the entire row container for the password field.
    const passwordContainer = prompt.container.querySelector('input[type="password"]').parentElement.parentElement
    passwordContainer.classList.add('hidden')
    prompt.open()

    function onContinue(formValues) {
      const modalBody = prompt.container.querySelector('div.dcp-modal-body')
      modalBody.innerText = 'Loading...'
      setTimeout(tryToLoadWallet.bind(this, formValues), 50)
    }

    function tryToLoadWallet(formValues) {
      const [fileInputElement, password, privateKey] = formValues
      let wallet
      try {
        // Checking if the keystore is coming from the file input or a private
        // key.
        if (fileInputElement.value && typeof keystore === 'string') {
          try {
            keystore = keystore.trim()
            window.sessionStorage.setItem('keystore', keystore)
            window.sessionStorage.setItem('passphrase', password)
            keystore = JSON.parse(keystore)
          } catch (error) {
            reject(Error('Unable to parse keystore file.'))
          }
          resolve(new KeystoreConstructor(keystore, password))
        } else {
          resolve(new KeystoreConstructor(privateKey))
        }
      } catch (error) {
        return reject(error)
      }
      resolve(new KeystoreConstructor(keystore))
    }

    function handleExitingUpload(modal) {
      reject(Error('Cancelled Wallet Upload.'))
    }

    function validateOnSubmit(event) {
      //event.preventDefault()
      const fileFormElements = prompt.container.querySelectorAll('.dcp-modal-body input[type="file"], .dcp-modal-body input[type="text"]')
      const filledInFileFormElements = Array.from(fileFormElements).filter(fileFormElement => fileFormElement.value !== '')
      if (filledInFileFormElements.length === 0) {
        const continueButton = prompt.container.querySelector('.dcp-modal-footer button.continue')
        continueButton.setCustomValidity('Please select a keystore file or enter a private key.')
        // The timeout ensures the validity is reported correctly once the main thread opens up.
        window.setTimeout(() => continueButton.reportValidity())
        return
      }
      prompt.continue()
    }
  })
}

wallet$$get_node = async function (argv, options) {
  return (await exports.load(options)).keystore
}

let getKeystore_platform = require('../dcp-env').switch({
  nodejs: wallet$$get_node,
  bravojs: wallet$$get_browser,
  'vanilla-web': wallet$$get_browser,
  default: typeof navigator === 'undefined' ? wallet$$get_node : wallet$$get_browser
})

async function get_internal(KeystoreConstructor, defaultName, getArguments) {
  let options, argv, ks

  if (getArguments[0] === false) {
    getArguments.shift()
    ksCache('reset')
  } else {
    let cached = ksCache('get', KeystoreConstructor, defaultName, getArguments)
    if (cached)
      return cached
  }
  
  if (getArguments.length === 1 && typeof getArguments[0] === 'object' && !Array.isArray(getArguments[0]))
    options = getArguments[0]
  else if (getArguments.length === 2 && typeof getArguments[1] === 'object')
    options = getArguments[1]
  else
    options = {}

  if (Array.isArray(getArguments[0])) {
    argv = getArguments[0]
    
    for (let i=0; i < argv.length; i++) {
      if (argv[i][0] !== '-')
        continue
      switch(argv[i][1]) {
      case '-':
        let match = argv[i].match(/^--private-key=(.*)/)
        if (match) {
          let key = match[1]
          return new exports.Keystore(key, '')
        }
        break;
        case 'p':
          options.checkEmpty = false;
          break;
        case 'i':
          options.name = argv[i+1]
        default:
          argv.splice(i, 2)
          i -= 2
          break;
      }
    }
  }
  
  if (!options.hasOwnProperty('name'))
    options.name = defaultName

  if (typeof getArguments[0] === 'string')
    options.name = getArguments[0]
  
  if (typeof getArguments[1] === 'string')
    options.name = getArguments[1]

  if (typeof argv !== 'undefined')
    argv = []

  options.KeystoreConstructor = KeystoreConstructor
  ks = await getKeystore_platform(argv, options)
  ksCache('set', KeystoreConstructor, defaultName, getArguments, ks)
  return ks
}

function ksCache(oper, ctor, name, getArgs, ks) {
  let key;
  let overrideKey;
      
  if (oper === 'reset' || typeof ksCache.storage === 'undefined') {
    debugging('cache') && console.debug('wallet - reset cache');
    ksCache.storage = {}
    if (oper === 'reset')
      return;
  }

  key         = [/*ctor.name,*/ name, JSON.stringify(getArgs)].join();
  overrideKey = [/*ctor.name,*/ name, JSON.stringify(true)].join();

  if (oper === "get") {
    let entry = ksCache.storage[overrideKey] || ksCache.storage[key];
    debugging('cache') && console.debug(`wallet - cache ${entry ? 'hit' : 'miss'} on ${ctor.name} as ${name}`);
    debugging('cache') && entry && console.debug(`wallet - cached entry is ${ctor.name} as ${name} => ${entry.address} (${entry.label})`);
    return entry;
  }

  if (oper === "set") {
    debugging('cache') && console.debug(`wallet - cached ${ctor.name} as ${name} => ${ks.address} (${ks.label})`, ks.address);
    return ksCache.storage[key] = ks
  }

  throw new Error(`Invalid operation: ${oper}`)
}

/** Return an identity keystore */
exports.getId = /*async*/ function wallet$$getId() {
  return get_internal(exports.IdKeystore, 'default', arguments);  /** @todo enable identity key support /wg nov 2019 */
  /** XXX @todo enable identity key support */
  return get_internal(exports.IdKeystore, 'identity', arguments)
}

/** Return an authorization keystore */
exports.get = /* async */ function wallet$$get() {
  return get_internal(exports.AuthKeystore, 'default', arguments)
}

/** Add a temporary Keystore (in RAM) to the current wallet, which can be retrieved later by wallet.get().
 *  @param      name            The name of the keystore (default is 'default')
 *  @param      arguments       The arguments to pass to get(), along with the 
 *                              specifiedname, to retrieve this keystore.  Specifying the value
 *                              true (or not specifying it at all) causes get to retrieve this keystore 
 *                              for any arguments, provided name matches.
 */
exports.add = function wallet$$add(ks, name, getArguments) {
  if (!name)
    name = 'default'
  ksCache("set", exports.AuthKeystore, name, getArguments || typeof getArguments === 'undefined', ks)
 }

/** Same as exports.add(), except for retrieval by wallet.getId() */    
exports.addId = function wallet$$add(ks, name, getArguments) {
  if (!name)
    name = 'identity'
  ksCache("set", exports.IdKeystore, name, getArguments || typeof getArguments === 'undefined', ks)
  /** XXX @todo decouple identity / bank account keystores /wg dec 2019 */
  exports.add(ks, 'default', getArguments);
}

exports.clear = function wallet$$clear() {
  ksCache('reset');
}

/**
 * This function behaves exactly the same as get(), except it returns a private key.
 *
 * @param   {object | string} param1 | param2 - Three accepted keys: name, which specifies the name of the keystore, dir, which specifies its directpry, 
 *                                              and checkEmpty, which controls weather or not unlock() will attempt an empty password on the keystore 
 *                                              before prompting the user.
 * @returns {Object}                          - AuthKeystore object
 */
exports.getPrivateKey = wallet$$getPrivateKey = async AuthKeystoreObj => {
  if (typeof AuthKeystoreObj == 'object') {
    //await AuthKeystoreObj.unlock()
    return await AuthKeystoreObj.getPrivateKey()
  } else {
    throw new TypeError('AuthKeystore Object')
  }
}

async function getVerifiedPassphrase() {
  let pp1, pp2;
  let ks = {}
  
  do {
    pp1 = await exports.passphrasePrompt('new key')
    pp2 = await exports.passphrasePrompt('new key again', true)
  } while(pp1 !== pp2)

  return pp1
}
                                  
/** Generate a new private key */
function generateNewPrivateKey() {
  return new exports.PrivateKey(require('ethereumjs-wallet').generate().getPrivateKey())
}

/**
 * This function takes an encrypted keystore and returns its address without decrypting it
 *
 * @param   {keystore}
 * @returns {Buffer}
 */
exports.getAddress = wallet$$getAddress = keystore => {
  return keystore.address
}

/** Accept json from any foreign ethereum wallet [that we understand] and a passphrase,
 *  returning the address and private key
 *
 *  @param      json            {string}        A JSON keystore, eg. ethereum V3 wallet
 *  @param      passphrase      {string}        The passphrase to unlock the keystore
 *  @returns    {object}                        An object having properties address and privateKey
 */
function parseForeignKeystore(json, passphrase) {
  const etherWallet = require('ethereumjs-wallet')
  let fks

  try {
    fks = etherWallet.fromV3(json, passphrase, { nonStrict: true })
  } catch(e) {
    try {
      fks = etherWallet.fromV1(json, passphrase)
    } catch(e) {
      try {
        fks = etherWallet.fromEthSale(json, passphrase)
      } catch(e) {
        return false
      }
    }
  }

  return { address: fks.address, privateKey: fks.getPrivateKey() }
}

/** This function tries to get the correct passphrase, by asking the user
 *  (via exports.passphrasePrompt) up to exports.maxPassphraseTries times.  
 *  The passphrase is tested by passing it into tryFn. If tryFn returns truey,
 *  that value is immediately returned.  If all tries are used and tryFn
 *  never returns truey, this function returns false.
 *
 *  @param   label {string}    label which is passed to exports.passphrasePrompt()
 *  @param   tryFn {function}  function which returns 
 */
async function passphraseTries(label, tryFn, skipCheckFn) {
  let res, pp
  
  for (let count = 0; count < exports.maxPassphraseTries; count++) {
    pp = await exports.passphrasePrompt(label, count > 0, skipCheckFn)
    res = tryFn(pp)
    if (res)
      return res
  }

  return false
}

/**
 * This method creates functions related to locking and unlocking Keystores which share a common set
 * of private variables and are added to the keystore as own property.
 *
 * @param       ks      {Object}        instance of keystore
 * @param       ew      {String}        json which represents an Ethereum V3 keystore [ethereum wallet]
 */
function KeystoreLockFunctionsFactory(ks, ew) {
  let privateKey
  let autoResetDuration
  let lockExpireTime, lockTimer
  let label = ks.label || ks.address

  function wallet$$Keystore$resetTimerState() {
    debugging('timer') && console.log(`wallet - reset lock timer state on ${label}`);
    clearTimeout(lockTimer)
    autoResetDuration = 0
    lockExpireTime = Date.now()
    lockTimer = undefined
  }

  wallet$$Keystore$resetTimerState()

  /* belt-and-suspenders (checkTimerState + setTimeout) used to decrease
   * the possibility that starving the event loop causes the private key
   * to be stored for longer than intended
   */
  function wallet$$Keystore$checkTimerState() {
    debugging('timer') && console.log(`wallet - checkTimerState on ${label}`);
    if (lockTimer && lockExpireTime <= Date.now()) {
      wallet$$Keystore$lock()
    }
  }
    
  /**
   * This method unlocks the Keystore object for either a single use or a given amount of time. When the 
   * object is unlocked, `getPrivateKey` and related operations do not need to ask for a passphrase, as 
   * the private key is "remembered" internally.  If time is not specified, the object will be immediately 
   * locked after the next call to `getPrivateKey`; otherwise, it will remain unlocked until the timer runs out.
   * 
   * When the unlock timer is running, calling `unlock` can have multiple effects. In all cases the current
   * running timer will be increased to lockTimerDuration if that will increase the time left.
   *   1. If the autoReset flag is true for this invocation, the autoReset duration is updated only if it
   *      will be increased. Thus autoReset is a sticky setting, it can be turned on but not off (except by locking)
   *   2. If the autoReset flag is false for this invocation, the current unlock duration is increased, but 
   *      the autoReset duration remains the same.
   *
   * 
   * @param  passphrase          {?string}  Passphrase to unlock keystore. If null|undefined, `passphrasePrompt` 
   *                                          function is invoked to solicit it from the user if the keystore
   *                                          is not already unlocked.
   * @param  lockTimerDuration  {number}   Number of seconds to keep keystore unlocked for.
   * @param  autoReset          {boolean}  Flag to determine if unlock timer should reset on subsequent calls.
   */
  async function wallet$$Keystore$unlock(passphrase, lockTimerDuration, autoReset) {
    let wasLockedAtBeginning = wallet$$Keystore$isLocked()
    let newExpireTime = Date.now() + (lockTimerDuration * 1000)
    
    if (wasLockedAtBeginning) {
      wasSet = await setPrivateKeyFromForeignKeystore(passphrase)
      if (wallet$$Keystore$isLocked()) throw new Error(`Could not unlock keystore '${label}'`)
      // If the keystore is unlocked but privateKey was not set by the above method, the settings
      // for `lockTimerDuration` and `autoReset` are now associated to bad credentials and shan't be
      // propogated.
      if (!wasSet) return
    }

    if (!wasLockedAtBeginning && increasingPermissivenessOfTimerState(newExpireTime, autoReset, lockTimerDuration)) {
      let canIncrease = await setPrivateKeyFromForeignKeystore(passphrase)
      if (!canIncrease) return
    }

    if (autoReset && lockTimerDuration > autoResetDuration)
      autoResetDuration = lockTimerDuration
      
    if (newExpireTime > lockExpireTime) {
      debugging('timer') && console.log(`wallet - setting new lock expire time on ${label} - ${lockTimerDuration * 1000}ms`);
      wallet$$Keystore$setTimerState(lockTimerDuration)
    } 

    if (wallet$$Keystore$isLocked())
      throw new Error('Keystore could not be unlocked.')
    
    debugging('locks') && console.log(`wallet - ${label} is unlocked`);
  }

  /**
   * Are we increasing the permissiveness of the timer state?
   */ 
  function increasingPermissivenessOfTimerState (newExpireTime, autoReset, lockDuration) {
    return autoReset ? lockDuration > autoResetDuration : newExpireTime > lockExpireTime 
  }

  /** Lock the Keystore */
  function wallet$$Keystore$lock() {
    let debugMsg = `wallet - lock ${label}`
    if (wallet$$Keystore$isUnlocked()) debugMsg += ' (was not locked)'
    debugging('locks') && console.log(debugMsg);
    
    privateKey = undefined
    wallet$$Keystore$resetTimerState()
  }

  async function wallet$$Keystore$getPrivateKey() {
    let pk

    wallet$$Keystore$checkTimerState()
    if (wallet$$Keystore$isLocked()) {
      await wallet$$Keystore$unlock()
      pk = privateKey
      wallet$$Keystore$lock()
    } else {
      pk = privateKey
    
      if (autoResetDuration > 0) {
        debugging('timer') && console.log(`wallet - auto reset new lock expire time on ${label} - ${autoResetDuration * 1000}ms`);
        wallet$$Keystore$setTimerState(autoResetDuration)
      } else if (!lockTimer) {
        wallet$$Keystore$lock()
      }
   }
    
    debugging('getPrivateKey') && console.log(`wallet - ${label} pk=${pk}`);
    return pk
  }

  function wallet$$Keystore$setTimerState(newDuration) {
    lockExpireTime = Date.now() + (newDuration * 1000)
    clearTimeout(lockTimer)
    lockTimer = setTimeout(wallet$$Keystore$lock, newDuration * 1000)
  }

  /**
   * @returns {boolean} Was this keystore successfully unlocked?
   */
  async function setPrivateKeyFromForeignKeystore(passphrase) {
    let fks
    if (passphrase !== undefined && passphrase !== null) {
      fks = parseForeignKeystore(ew, passphrase)
    } else {
      fks = parseForeignKeystore(ew, '')
      if (!fks) {
        fks = await passphraseTries(label, 
                                    function wallet$$Keystore$Constructor$unlock_tryFn(passphrase) {
                                      return parseForeignKeystore(ew, passphrase)
                                    },
                                    function wallet$$Keystore$Constructor$unlock_skipCheckFn() {
                                      return !!privateKey
                                    })
      }
    }

    if (!fks)
      return false

    privateKey = new exports.PrivateKey(fks.privateKey)
    return true
  }

  function wallet$$Keystore$isUnlocked() {
    return !!privateKey
  }

  function wallet$$Keystore$isLocked() {
    return !wallet$$Keystore$isUnlocked()
  }
  
  ks.isLocked      = wallet$$Keystore$isLocked
  ks.isUnlocked    = wallet$$Keystore$isUnlocked
  ks.unlock        = wallet$$Keystore$unlock
  ks.lock          = wallet$$Keystore$lock
  ks.getPrivateKey = wallet$$Keystore$getPrivateKey
}

/**
 * Class that represents a keystore for use with DCP. This is essentially an encrypted data structure that stores a
 * private key and a known address.  The private key is used by one party to sign transactions over DCP
 * in a way which allows the other party to know (within cryptographic certainty) that the first party
 * knows the private key, and hence is authorized to move funds associated with the Address.
 *
 * Since this constructor might have to prompt a user for a password, we will need async behavior. This will be 
 * accomplished by returning a promise (a Wes idea). All async operations will take place in that promise.
 *
 * @constructor
 *
 * form 1: 
 * @param none
 * @returns a promise resolving to a keystore object with a randomly generated privateKey and an address
 * that corresponds to it. This form will prompt the user for a password to encrypt itself with.
 *
 * form 2: 
 * @param privateKey {string|object}
 * @returns a promise resolving to a keystore object with the provided privateKey and an address that 
 * corresponds to it. This form will prompt the user for a password to encrypt itself with.
 *
 * form 3: 
 * @param privateKey {string|object}
 * @param passphrase {string}
 * @returns a promise resolving to a keystore object with the provided privateKey and an address that 
 * corresponds to it, encrypted with the provided passphrase.
 *
 * form 4: 
 * @param privateKey {third-party keystore object}
 * @param passphrase {string}
 * @returns a promise resolving to a keystore object parsed from the third party object. If the third party
 * object cannot be parsed the promise will reject.
 *
 * form 5:
 * @param JsonString {string} - a JSON encoded keystore of some sort
 * @param passphrase {string}
 * @returns a promise resolving to a keystore object parsed from the JSON string. If the JSON or resulting
 * object cannot be parsed the promise will reject.
 *
 * form 6:
 * @param privateKey {third-party keystore object}
 * @returns a promise resolving to a keystore object parsed from the third party object. If the third party
 * object cannot be parsed the promise will reject. The user will be prompted for a passphrase.
 *
 * form 7: 
 * @param null, passphrase
 * @returns a promise resolving to a keystore object with a randomly generated privateKey and an address
 * that corresponds to it. This form will used the supplied passphrase to encrypt the private key.
 */
exports.Keystore = function wallet$$Keystore() {
  if (arguments[0] === _ConstructOnly)
    return;

  let ctor = {}
  let ew  /* ethereum wallet: (json) backing store for encrypted private keys */
  
  this.toJSON = function wallet$$Keystore$toJSON() {
    if (ctor.ks)
      return ctor.ks
    return JSON.parse(ew)
  }

  return (async () => {
    const etherWallet = require('ethereumjs-wallet')
    const etherUtil   = require('ethereumjs-util')
    const isBuffer    = require('is-buffer');

    if (arguments.length === 2)
      ctor.passphrase = arguments[1]                              /* all forms which accept passphrases */
    
    if (arguments.length === 0) {
      ctor.privateKey = generateNewPrivateKey()                   /* form 1 */
      ctor.passphrase = await getVerifiedPassphrase()
    }
    else if (arguments[0] === null)                               /* form 7 */
      ctor.privateKey = generateNewPrivateKey()
    else if (exports.PrivateKey(arguments[0], true)) {            /* form 2, 3 */
      if (ctor.passphrase === false)
        ctor.passphrase = '';
      if (typeof ctor.passphrase === 'undefined')
        ctor.passphrase = await getVerifiedPassphrase()        
      ctor.privateKey = new exports.PrivateKey(arguments[0])
    }
    else if (arguments[0] instanceof exports.Keystore) {          /* form 6 */
      ctor.ks = arguments[0]
      this.address = ctor.ks.address
    }
    else if (typeof arguments[0] === 'string' ||
             typeof arguments[0] === 'object') {                  /* form 4, 5 */
      let jsonO

      if (typeof arguments[0] === 'object') {     /* form 4 */
        ew = JSON.stringify(arguments[0])
        jsonO = arguments[0]
      } else {                                    /* form 5 */
        ew = arguments[0]
        jsonO = JSON.parse(ew.trim())
      }
      if (jsonO.address)
        this.address = new exports.Address(jsonO.address)
      else {
        let fks
        
        if (ctor.passphrase)
          fks = parseForeignKeystore(ew, ctor.passphrase)
        else {
          fks = parseForeignKeystore(ew, '')
          if (!fks)
            fks = await passphraseTries(jsonO.label || jsonO.id,
                                        function wallet$$Keystore$Constructor$unlock_tryFn(passphrase) {
                                          parseForeignKeystore(ew, passphrase)
                                        })
        }

        if (!fks)
          throw new Error('Could not unlock foreign keystore ${jsonO.label || jsonO.id} to determine public address')
        this.address = fks.address
      }
      if (jsonO.label)
        this.label = jsonO.label
    }

    /* By the bottom of this funnel, we have the 
     * - defined this.address
     * - PrivateKey or a ethereum wallet (JSON string)
     * - passphrase or '' if making a Keystore from a PrivateKey
     *
     * Now we make sure that we can find the private key by unlocking ew, 
     * have an address, and forget the private key
     */
    if (!ew) {
      let buf = etherUtil.toBuffer(ctor.privateKey.toString())
      let i = etherWallet.fromPrivateKey(buf)
      ew = i.toV3String(etherUtil.toBuffer(ctor.passphrase || ''), { n: 1024 })
    }
    if (ctor.privateKey) {
      this.address = ctor.privateKey.toAddress()
    }
    delete ctor.privateKey
    delete ctor.passphrase
    KeystoreLockFunctionsFactory(this, ew)

    return this
  })()
}
           
// these 3 classes extend the Keystore class, right now, they don't have any extra funcitonality, but this will be expanded upon in later specs

exports.IdKeystore = function walletAPI$$IdKeystore(privateKey, passphrase) { 
  return this.constructor.prototype.constructor.apply(this, arguments)
}

exports.BankAccountKeystore = function walletAPI$$BankAccountKeystore(privateKey, passphrase) {
  return this.constructor.prototype.constructor.apply(this, arguments)
}

exports.AuthKeystore = function walletAPI$$AuthKeystore(privateKey, passphrase) {
  return this.constructor.prototype.constructor.apply(this, arguments)
}

exports.IdKeystore.prototype = new exports.Keystore(_ConstructOnly)
exports.BankAccountKeystore.prototype = new exports.Keystore(_ConstructOnly)
exports.AuthKeystore.prototype = new exports.Keystore(_ConstructOnly)

exports.maxPassphraseTries = 5

