/** 2
 *  @file       workerBootstrap.js
 *              Copyright (c) 2018, Kings Distributed Systems, Ltd.  All Rights Reserved.
 *
 *              Simple bootstrap code which lets us eval code in an empty worker, so that
 *              we can load bravo.js into the worker.
 *
 *  @author     Wes Garland, wes@sparc.network
 *  @date       May 2018
 *  @module     WorkerBootstrap
 */
/* globals window, self, global */

var global = this;


// aggregated from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects#Reflection
var whitelist = new Set([
  '__proto__',
  '_console',
  'addEventListener',
  'applyWhitelist',
  'Array',
  'ArrayBuffer',
  'AsyncFunction',
  'atob',
  'Atomics',
  'Boolean',
  'Blob',
  'bravojs',
  'btoa',
  'clearInterval',
  'clearTimeout',
  'console',
  'constructor',
  'DataView',
  'Date',
  'decodeURI',
  'decodeURIComponent',
  'encodeURI',
  'encodeURIComponent',
  'Error',
  'escape',
  'eval',
  'EvalError',
  'File',
  'FileReader',
  'Float32Array',
  'Float64Array',
  'Function',
  'global',
  'Generator',
  'GeneratorFunction',
  'Headers',
  'Infinity',
  'Int16Array',
  'Int32Array',
  'Int8Array',
  'isFinite',
  'isNaN',
  'JSON',
  'Map',
  'Math',
  'module',
  'NaN',
  'navigator',
  'null',
  'Number',
  'Object',
  'OffscreenCanvas',
  'onerror',
  'onmessage',
  'parseFloat',
  'parseInt',
  'performance',
  'postMessage',
  'Promise',
  'propertyIsEnumerable',
  'Proxy',
  'RangeError',
  'ReferenceError',
  'RegExp',
  'removeEventListener',
  'requestAnimationFrame',
  'require',
  'Response',
  'self',
  'Set',
  'setInterval',
  'setTimeout',
  'sleep',
  'String',
  'Symbol',
  'SyntaxError',
  'TextDecoder',
  'TextEncoder',
  'toLocaleString',
  'toString',
  'TypeError',
  'URIError',
  'URL',
  'Uint16Array',
  'Uint32Array',
  'Uint8Array',
  'Uint8ClampedArray',
  'undefined',
  'unescape',
  'valueOf',
  'WeakMap',
  'WeakSet',
  'WebAssembly',
  'WebGL2RenderingContext',
  'WebGLTexture',

  // Our own symbols
  'progress',
  'work',
])

// Set values to true to disallow access to symbols
var blacklist = {
  OffscreenCanvas: false,
}

const blacklistRequirements = {
  OffscreenCanvas: "environment.offscreenCanvas"
}

/**
 * Applies a whitelist and a blacklist of properties to an object. After this function, if someone tries
 * to access non-whitelisted or blacklisted properties, a warning is logged and it will return undefined.
 *
 * @param {object} obj - The object, which will have the whitelist applied to its properties.
 * @param {Set} whitelist - A set of properties to allow people to access.
 * @param {Set} blacklist - An object of property names mapping to booleans to indicate whether access is allowed or not.
 * @param {Set} blacklistRequirements - An object of property names mapping requirement path strings, used to print useful warnings.
 */
function applyAccessLists(obj, whitelist, blacklist={}, blacklistRequirements={}) {
  if (!obj) { return }
  Object.getOwnPropertyNames(obj).forEach(function(prop) {
    if (Object.getOwnPropertyDescriptor(obj, prop).configurable) {
      if (!whitelist.has(prop)) {
        let isSet = false
        let propValue
        Object.defineProperty(obj, prop, {
          get: function () {
            if (isSet) {
              return propValue
            } else {
              const error = new Error('Security Warning: cannot access ' + prop + ', returning undefined.')
              console.warn('Worker global whitelist error: cannot access ' + prop, error.stack)
              postMessage({
                request: 'accessListViolation',
                message: error.message
              });
              return undefined
            }
          },
          set: function (value) {
            propValue = value
            isSet = true
          },
          configurable: false
        })
      } else if (prop in blacklist) {
        let isSet = false;
        let blacklisted = blacklist[prop];
        let requirement = blacklistRequirements[prop];
        let propValue = obj[prop];
        Object.defineProperty(obj, prop, {
          get: function () {
            if (blacklisted && !isSet) {
              const error = new Error('Cannot access ' + prop + ', returning undefined.')
              console.warn('Worker global blacklist error: Cannot access ' + prop + ' without specifying ' + requirement + '=true in your job\'s requirements object.' , error.stack)
              postMessage({
                request: 'accessListViolation',
                message: error.message
              });
              return undefined
            } else {
              return propValue
            }
          },
          set: function (value) {
            propValue = value
            isSet = true
          },
          configurable: false
        })
      }
    }
  })
}

/**
 * Applies the whitelist and blacklist to all global scopes.
 * This must be called after the requirements are assigned to the sandbox
 * so that the blacklist is accessible to modify w/o adding it to the whitelist.
 */
function applyAllAccessLists() {
  // We need to apply the access lists to global, global.__proto__, and global.__proto__.__proto__,
  // because there's networking-accessing functions inside global.__proto__.__proto__, like fetch.
  //
  // If we're in a robust environment (node, browser, WebWorker, basically anything but v8),
  // then we have to climb the prototype chain and apply the whitelist there, but we have to stop
  // before we whitelist Object's properties

  // Save them in scope because they'll get hidden by the whitelist
  let _whitelist = whitelist;
  let _blacklist = blacklist;
  let _blacklistRequirements = blacklistRequirements;
  let _applyAccessLists = applyAccessLists;
  for (let g = global; g.__proto__ && (g.__proto__ !== Object); g = g.__proto__) {
    _applyAccessLists(g, _whitelist, _blacklist, _blacklistRequirements);
  }

  if (typeof navigator === 'undefined') {
    navigator = { userAgent: 'not a browser' }
  } else {
    // We also want to whitelist certain parts of navigator, but not others.
    navWhitelist = new Set(['userAgent'])
    _applyAccessLists(navigator.__proto__, navWhitelist)
  }
}

function getCapabilities() {
  let offscreenCanvas = false
  , bigTexture4096 = false
  , bigTexture8192 = false
  , bigTexture16384 = false
  , bigTexture32768 = false
  , es7 = false
  , spidermonkey = false;

  let fdlibmFlag = true;
  let inputFdlibm = [
    7.330382858376182753090688493103,
    9.424777960769379347993890405633,
    14.660765716752370835251895186957,
    16.755160819145565653798257699236,
    7.330382858376180000000000000000
  ];

  let outpuFdlibm = [
    1525.965888988748247356852516531944,
    12391.647807916691817808896303176880,
    2328571.894357232376933097839355468750,
    18909231.863243132829666137695312500000,
    1525.965888988744154630694538354874
  ];

  offscreenCanvas = !!(typeof OffscreenCanvas !== 'undefined' && new OffscreenCanvas(1, 1));

  if (offscreenCanvas) {
    let canvas = new OffscreenCanvas(1, 1);
    let gl = canvas.getContext('webgl');
    let textureSize = gl.getParameter(gl.MAX_TEXTURE_SIZE);

    // Destroy the WebGL context, from https://www.khronos.org/registry/webgl/extensions/WEBGL_lose_context/
    // "This is the recommended mechanism for applications to programmatically halt their use of the WebGL API."
    gl.getExtension('WEBGL_lose_context').loseContext();

    bigTexture4096 = textureSize >= 4096
    bigTexture8192 = textureSize >= 8192
    bigTexture16384 = textureSize >= 16384
    bigTexture32768 = textureSize >= 32768
  }

  try {
    for (let i = 0; i < inputFdlibm.length; i++) {
      if (Math.exp(inputFdlibm[i]).toFixed(30) !== outpuFdlibm[i].toFixed(30)) {
        fdlibmFlag = false
        break;
      }
    }
  } catch (err) {
    fdlibmFlag = false
  }

  return {
    engine: {
      es7,
      spidermonkey,
    },
    environment: {
      offscreenCanvas,
      fdlibm: fdlibmFlag,
    },
    details: {
      offscreenCanvas: {
        bigTexture4096,
        bigTexture8192,
        bigTexture16384,
        bigTexture32768,
      },
    },
  }
}

/* Polyfill section of workerBootstrap */

// At time of writing, Chrome defines requestAnimationFrame inside web workers, but
// Firefox doesn't.
if (typeof requestAnimationFrame == 'undefined') {
  global.requestAnimationFrame = callback => setTimeout(callback, 0)
}

if (typeof OffscreenCanvas !== 'undefined') {

  // This deals with Firefox bug 1529995, which causes the tab to crash if fenceSync is called.
  if (navigator.userAgent.indexOf('Firefox') >= 0) {
    new OffscreenCanvas(640, 480).getContext('webgl2').__proto__.fenceSync = null
    // Note: We can't just do the following, since WebGL2RenderingContext isn't defined
    // in Firefox until the first webgl2 context is created.
    // WebGL2RenderingContext.prototype.fenceSync = undefined
  }

  // Make it so that if getContext throws on a given type of context, return null
  // instead of throwing an exception. This replicates Chrome's behaviour.
  OffscreenCanvas.prototype.oldGetContext = OffscreenCanvas.prototype.getContext
  OffscreenCanvas.prototype.getContext = function getContextPolyfill(type) {
    try {
      return this.oldGetContext(type)
    } catch (e) {
      return null
    }
  }
}

bravojs = {
  url: '/bravojs/bravo.js',
  mainModuleDir: '.'
}
if (bravojs !== self.bravojs) { self.bravojs = bravojs }

(function (addEventListener, removeEventListener, postMessage) {
  let lastProgress = 0;

  let workerBootstrap$messageHandler
  addEventListener('message', workerBootstrap$messageHandler = async (event) => {
    try {
      var indirectEval = eval // eslint-disable-line
      if (event.data.request === 'eval') {
        try {
          let result = await indirectEval(event.data.data, event.data.filename)
          postMessage({
            request: 'evalResult',
            data: result
          })
        } catch (error) {
          postMessage({
            request: 'uncaughtException',
            error: {
              name: error.name,
              message: error.message,
              stack: error.stack.replace(
                /data:application\/javascript.*?:/g,
                'eval:'
              ),
            }
          })
        }
      } else if (event.data.request === 'applyRequirements') {
        // This event is fired when the worker is initialized with generator requirements,
        // apply restrictions to the environment based on the requirements.
        // Assume the scheduler gave us a nicely-shaped req object.
        const requirements = event.data.requirements;
        blacklist.OffscreenCanvas = !requirements.environment.offscreenCanvas;
        applyAllAccessLists();

        postMessage({ request: 'applyRequirementsDone' });
      } else if (event.data.request === 'resetState') {
        // This event is fired when the web worker is about to be reused with another slice
        lastProgress = 0;
        postMessage({ request: 'resetStateDone' });
      } else if (event.data.request === 'checkCapabilities') {
        // This event is fired when the web worker checks capabilities.
        let result = getCapabilities();
        postMessage({
          request: 'checkCapabilitiesDone',
          data: result
        })
      }
    } catch (error) {
      postMessage({
        request: 'uncaughtException',
        error: {
          name: error.name,
          message: error.message,
          stack: error.stack,
        }
      })
    }
  })

  // Save the old "real" console before it gets replaced
  // let _console = console;

  const emitNoProgress = (message) => {
    lastProgress = null;
    postMessage({
      request: 'noProgress',
      message
    });
  }

  self.progress = function workerBootstrap$progress(value) {
    // lastProgress is set to null when noProgress is emitted,
    // prevents multiple noProgress events from firing
    if (lastProgress === null) return false;

    let progress, isIndeterminate = false;
    if (value === undefined) {
      progress = lastProgress || 0;
      isIndeterminate = true;
    } else {
      progress = parseFloat(value);

      if (Number.isNaN(progress)) {
        isIndeterminate = true;
      } else {
        if (!(typeof value === 'string' && value.endsWith('%'))) {
          // if the progres value isn't a string ending with % then multiply it by 100
          progress *= 100;
        }
      }
    }

    if (progress < 0 || progress > 100) {
      emitNoProgress(`Progress out of bounds: ${progress.toFixed(1)}%, last: ${lastProgress.toFixed(1)}%`);
      return false;
    } else if (progress < lastProgress) {
      // Nerf reverse progress error, mark as indeterminate // RR Jan 2020
      progress = lastProgress;
      isIndeterminate = true;
    }

    lastProgress = progress;

    postMessage({
      request: 'progress',
      progress,
      isIndeterminate
    });
    return true;
  }

  function workerBootstrap$work$emit(eventName, value) {
    if (typeof eventName !== 'string') {
      throw new Error(`Event name passed to work.emit must be a string, not ${eventName}.`);
    }

    postMessage({
      request: 'emitEvent',
      payload: value,
      eventName,
    });
  }

  self.work = { emit: workerBootstrap$work$emit };
  // self.dcp = {
  //   work: { emit: workerBootstrap$work$emit },
  // }

  function workerBootstrap$console(level, ...args) {
    let message = args.map(a =>
      (typeof a === 'string'? a : JSON.stringify(a))
    ).join(' ');

    postMessage({
      request: 'console',
      payload: {
        level,
        message,
        fileName: undefined,
        lineNumber: undefined,
      }
    });
  }

  self.console = {
    log:    workerBootstrap$console.bind(null, 'log'),
    debug:  workerBootstrap$console.bind(null, 'debug'),
    info:   workerBootstrap$console.bind(null, 'info'),
    warn:   workerBootstrap$console.bind(null, 'warn'),
    error:  workerBootstrap$console.bind(null, 'error'),
  };
})(self.addEventListener, self.removeEventListener, self.postMessage)

'evaluated workerBootstrap.js'
