/* sync: dcp src/debugging.js 3fc1ea0ea41b977cb490432edbb5871c90b6b2a2 */

/**
 *  @file       debugging.js
 *              Utility functions for improving debug output, etc, in DCP JavaScript programs.
 *
 *  @author     Wes Garland, wes@kingsds.network
 *  @date       Nov 2019
 */

/**
 * Declare a new debugging scope.  Each scope can watch a separate watchObject to determine
 * whether or not we are currently in debug mode.  Every scope watches the environment 
 * variable 'DCP_DEBUG'. 
 *
 * The debugging scope is queried by invoking the returned function, which has an optional
 * 'subsystem' argument.  When the query is made without the subsystem argument, the returned
 * function returns true if any subsystem in the scope is being debugged.  Otherwise, we return
 * true when the current debugging operations matches the debugging scope and specified subsystem.
 *
 * The watch object and the DCP_DEBUG environment variable are re-evaluated every time the 
 * returned function is called, however the results of matching for a given scope/subsystem are 
 * cached for speed.
 *
 * @param      {string}     scopeName      The name of the debugging scope (eg. 'compute')
 * @param      {object}     watchObject    An optional object whose `debug` property is watched
 *                                         to determine what debugging operation we are performing.
 * @returns    {function}   A function which returns boolean which is only true when the current
 *                          debugging operation describes both the current scope and subsystem.
 *
 * @example - logs when watchObject.debug or DCP_DEBUG='compute' or '*'
 * debugging = require('debugging').scope('compute');
 * debugging() && console.log('hello')
 *
 * @example - logs when watchObject.debug or DCP_DEBUG='compute-for' or '*'
 * debugging = require('debugging').scope('compute');
 * debugging('for') && console.log('hello')
 */
exports.scope = function debugging$$scope(scopeName, watchObject) {

  if (!scopeName)
    throw new Error('Must specify scopeName!');
  if (!watchObject)
    watchObject = {};
  
  function debugging(subsystem) {
    const env = process.env.DCP_DEBUG;
    const debugSyms = (env || '') + (env && watchObject.debug ? ',' : '') + (watchObject.debug || '');
    let result = false;
    var re;
    
    if (debugging.cacheValidity !== debugSyms) {
      debugging.cache = {};
      debugging.cacheValidity = debugSyms;
    }
    if (debugging.cache[subsystem])
      return debugging.cache[subsystem];

    loop: for (let debugSym of debugSyms.split(',')) {

      if (typeof subsystem === 'undefined') {
        if (typeof debugSym === 'boolean' && debugSym) {
          result = debugSym;
          break;
        }
        if (debugSym === scopeName) {
          result = true;
          break;
        }
      }
      
      switch(debugSym) {
        case '*': case 'all':                    /* everything */
        case scopeName + '-all' :                /* everything for this scope */
        case subsystem:                          /* everything for this subsystem */
        case scopeName + '-' + subsystem:        /* things for this scope and subsystem */
          result = true;
          break loop;
      }

      re = new RegExp('^(' + debugSym + ')$');
      if (re.test(scopeName) || re.test(subsystem) || re.test(scopeName + '-' + subsystem)) {
        result = true;
        break;
      }
    }
    
    debugging.cache[subsystem] = result;
    return result;
  }

  return debugging;
}
