/**
 *  @file   rtLink.js  A node module which handles runtime resolution of
 *                     module identifiers in a way which allows DCP code
 *                     to run from both the git repo and the installed
 *                     library locations tranparently.
 *
 *                     It accomplishes this by managing search paths during
 *                     node module resolution for modules under the dcp/
 *                     namespace.
 *
 *                     This module should be initialized before any other modules
 *                     are loaded.  It uses its own location inside node_modules/
 *                     to determine the correct running mode and to locate the 
 *                     local-config.json which points to the libraries it helps 
 *                     load.
 *
 *                     Non-source tree mode can be forced even when invoked from the
 *                     source tree by setting environment variable DCP_RTLINK_ASSUME_DIST_TREE.
 *
 *  @author Matthew Palma, mpalma@sparc.network; Wes Garland wes@kingsds.network
 *  @date   June 2018, Nov 2019
 */
const path = require('path')
const fs = require('fs')
const process = require('process')
const internalSym = {}
const debugging = require('./debugging').scope('rtlink', exports);
const dcpMagicPathPrefix = 'dcp/';
let   searchPath;
let   transpileDir;

var usingLegacySearch = false;
var newPaths;

/** Add rtLink$$getSearchPath() to the NODE_MODULES path.
 *  @deprecated
 */
exports.link = function rtLink$$link (paths, extraPaths, devExtraPaths) {
  let isWin = process.platform === 'win32';

  console.warn('Warning: using deprecated API rtLink$$link');
  debugging('environment') && console.log('rtlink isWin=', isWin)
  
  if (process.env.NODE_PATH)
    process.env.NODE_PATH += isWin ? ';' : ':';
  else
    process.env.NODE_PATH = ''
  process.env.NODE_PATH += exports.getSearchPath(paths, extraPaths, devExtraPaths).join(isWin ? ';' : ':');
  
  require('module').Module._initPaths()
  debugging() && console.log('rtlink NODE_PATH=', process.env.NODE_PATH);

  usingLegacySearch = true;
  exports.init(); /* capture new-style dcp/ namespace modules */
}

/**
 * Accepts an array and appends the correct paths for global module
 * resolution of DCP modules.
 *
 * @param {array} paths - this should be module.paths of the caller (deprecated)
 * @param {array} extraPaths - extra paths to append before the default systems paths
 * @param {array} devExtraPaths - extra paths to append before the default systems paths when running from the source tree
 *
 * @note        Any extra paths specified should not use relative module identifiers,
 *              as these will be resolved differently by different modules using the
 *              link() function if they are in different directories.
 */
exports.getSearchPath = function rtLink$$getSearchPath (paths, extraPaths, devExtraPaths) {
  /* Insert the new paths ahead of the OS-defined paths in the existing paths array */
  function applyNewPaths (paths) {
    if (!paths)
      return
    for (let i = newPaths.length; i > 0; i--) {
      if (paths.indexOf(newPaths[i - 1]) === -1) {
        paths.unshift(newPaths[i - 1]);
      }
    }
  }

  if (newPaths) {
    applyNewPaths(paths) /* re-use original extraPaths etc */
    return newPaths;
  }

  if (!fs.existsSync(exports.runningLocation)) {
    throw new Error('Could not determine running location!') /* should be impossible */
  }

  // for relatively pathed files (non-transpiled)
  newPaths = [
    path.resolve(path.dirname(exports.programName) + '/../lib'),
    path.join(exports.runningLocation, '/lib')
  ]

  if (exports.isSourceTree) {
    /* Add directories which are equivalent to /var/dcp/lib from a module path POV,
     * but live in different places in the source tree -- i.e. we are working out
     * of the source tree.
     */
    newPaths = newPaths.concat([
      path.join(exports.runningLocation, 'ormdb'),
      path.join(exports.runningLocation, 'src/node'),
    ])

    /* Try the installed (dist/) lib tree rather than the src/ files,
     * so that we pick up the transpiled modules. Try src/ next
     * to pick up the non-ESM modules in there that don't get installed
     */
    newPaths.push(path.join(exports.installLocation, '/lib'))
    newPaths.push(path.join(exports.runningLocation, '/src'))
    if (typeof devExtraPaths !== 'undefined')
      newPaths = newPaths.concat(devExtraPaths);
  }

  if (typeof extraPaths !== 'undefined') { newPaths = newPaths.concat(extraPaths) }

  applyNewPaths(paths)
  return newPaths;
}

const locatedModules = {}; /**< Cache modules which have been located with locateModule */
/** 
 * Locate a module which is under the magic path prefix.
 * Slightly different rules apply, depending on how rtlink() was initialized.  The intention is
 * for require('dcp/X') to work the same everywhere (webpack, node, dcp-client), finding the 
 * correct version of X to run.  This can be in different directories, depending on if we're
 * running in the source tree or in the dist hierarchy.
 *
 * @param       moduleIdentifer         {string}        argument to require() - should not
 *                                                      have filename extension!
 * @param       locateTranspiled        {boolean}       false to not try and locate a transpiled
 *                                                      version. This only happens when operating
 *                                                      out of the source tree.
 * @returns     {string} filename or false 
 */
function locateModule(moduleIdentifier, locateTranspiled) {
  let dir;
  let found = false;
  let searchModuleIdentifier = moduleIdentifier.slice(dcpMagicPathPrefix.length);

  if (locatedModules.hasOwnProperty(moduleIdentifier))
    return locatedModules[moduleIdentifier];

  for (dir of searchPath) {
    let filename = path.join(dir, searchModuleIdentifier + '.js');
    debugging('search') && console.log(`rtlink - searching for ${searchModuleIdentifier} as ${filename}`);
    if (fs.existsSync(filename)) {
      found = path.resolve(filename);
      break;
    }

    filename = path.join(dir, searchModuleIdentifier, 'index.js');
    debugging('search') && console.log(`rtlink - searching for ${searchModuleIdentifier} as ${filename}`);
    if (fs.existsSync(filename)) {
      found = path.resolve(filename);
      break;
    }
  }

  if (found && locateTranspiled !== false && !usingLegacySearch && exports.isSourceTree) {
    debugging('search') && console.log(`rtlink - checking for transpiled equivalent of ${searchModuleIdentifier} in ${transpileDir}`);

    let buffer = Buffer.allocUnsafe(128);
    let signature = Buffer.from('"__esModule"', 'utf-8');
    let testFilename = path.join(transpileDir, searchModuleIdentifier + '.js');
    let file;

    try {
      file = fs.openSync(testFilename);
      if (fs.readSync(file, buffer, 0, buffer.length, 0) >= signature.length) {
        if (buffer.indexOf(signature) !== -1) {
          debugging('search') || debugging('transpile') && console.log(`rtlink - using transpiled ${searchModuleIdentifier}`);
          if (fs.statSync(found).ctimeMs > fs.statSync(testFilename).ctimeMs) {
            console.warn(` * dcp-rtlink warning: stale transpilation detected for module ${moduleIdentifier}`);
            debugging('transpile') && console.log(` > source: ${found}, ${fs.statSync(found).ctimeMs}`)
            debugging('transpile') && console.log(` > babel:  ${testFilename}, ${fs.statSync(testFilename).ctimeMs}`)
          }
          found = testFilename;
        }
      }
    } catch(e) {
      if (e.code !== 'ENOENT')
        throw e;
    } finally {
      if (file)
        fs.closeSync(file);
    }
  }

  debugging('search') && found && console.log('rtlink - module found at', found);
  debugging('search') && !found && console.log(`rtlink - did not find module ${moduleIdentifier}`);

  locatedModules[moduleIdentifier] = found;
  return found;
}

/** Initialize the module; causes module-system hooks to "know" about the magical dcp/ namespace. */
exports.init = function rtLink$$init() {
  const moduleSystem = require('module');
  const resolveFilenamePrevious = moduleSystem._resolveFilename;
  let   distLib = path.join(exports.installLocation, 'lib');
   
  if (exports.isSourceTree) {
    searchPath = [
      path.join(exports.runningLocation, 'src'),
      path.join(exports.runningLocation, 'ormdb'),
      path.join(exports.runningLocation, 'src', 'node'),
    ];
  } else {
    searchPath = [ distLib ];
  }
  transpileDir = distLib;

  if (usingLegacySearch) {
    debugging('search') && console.log('rtlink - using legacy search');
    searchPath = exports.getSearchPath();
  }
  
  function rtLink$$init$resolveFilenameShim(moduleIdentifier) { 
    let filename;
    
    if (moduleIdentifier.startsWith(dcpMagicPathPrefix)) {
      filename = locateModule(moduleIdentifier);
      if (filename)
        return filename;
    }
        
    return resolveFilenamePrevious.apply(null, arguments)
  }

  moduleSystem._resolveFilename = rtLink$$init$resolveFilenameShim;
  delete exports.init;

  exports.link = function() {
    console.warn('Warning: using deprecated API rtLink$$link from library', new Error().stack.split('\n')[2]);
  }

  require('dcp/dcp-env').setPlatform('nodejs');
}

/* Determine where the node_modules/.. directory is that the caller used to find this module. */
let myTail = path.join(path.basename(path.dirname(module.filename)), path.basename(module.filename))
for (let dir of module.parent.paths) {
  let fullPath;
  let nodeModulesDirDetails = path.parse(dir);
  if (nodeModulesDirDetails.name !== 'node_modules')
    continue;

  fullPath = path.resolve(dir, myTail)
  debugging('search') && console.log('rtlink - checking', fullPath, fs.existsSync(fullPath));
  if (fs.existsSync(fullPath) && fs.realpathSync(fullPath) === module.filename) {
    exports.runningLocation = nodeModulesDirDetails.dir;
    break;
  }
}

if (!exports.runningLocation) {
  throw new Error('Could not determine rtLink.runningLocation!')
}

exports.installLocation = JSON.parse(fs.readFileSync(path.join(module.exports.runningLocation, 'etc', 'local-config.json'), 'utf-8')).dcp_root
exports.isSourceTree = !(path.resolve(exports.runningLocation) === path.resolve(exports.installLocation) || process.env.DCP_RTLINK_ASSUME_DIST_TREE);
exports.programName = process.argv[1] || '{node repl}'

if (debugging('environment')) {
  console.log('rtlink isSourceTree=', exports.isSourceTree)
  console.log('rtlink programName=', exports.programName)
}

if (debugging('paths')) {
  console.log('rtlink runningLocation=', exports.runningLocation)
  console.log('rtlink installLocation=', exports.installLocation)
}

/** Custom resolver for webpack (4.x or better) to understand the dcp/ module identifier nomeclature */
exports.WebpackResolve = function dcpRtlink$$WebpackResolve$(source, target) {
  this.source = source || 'resolve';
  this.target = target || 'resolve';
}

exports.WebpackResolve.prototype.apply = function dcpRtlink$$webpackResolvePlugin$apply(resolver) {
  var target = resolver.ensureHook(this.target);
  resolver.getHook(this.source).tapAsync('dcp-rtlink resolver', function(request, resolveContext, callback) {
    if (request.request.startsWith(dcpMagicPathPrefix)) {
      var obj = Object.assign({}, request, {
        request: locateModule(request.request, false) || ('dcp-rtlink/not-found/' + request.request)
      });
      return resolver.doResolve(target, obj, null, resolveContext, callback);
    }
    callback();
  });
}

